"""Test suite for basic backend simulation functionalities.
"""

# external _imports
from typing import Union
import numpy as np
import pytest

# pyrates internal _imports
from pyrates import integrate

# meta infos
__author__ = "Richard Gast, Daniel Rose"
__status__ = "Development"

###########
# Utility #
###########

# set backends to run the tests for
backends = ['default']

# set accuracy for all tests
accuracy = 1e-4


def setup_module():
    print("\n")
    print("==================================================")
    print("| Test Suite: Backend Simulation Functionalities |")
    print("==================================================")


def nmrse(x: np.ndarray,
          y: np.ndarray
          ) -> Union[float, np.ndarray]:
    """Calculates the normalized root mean squared error of two vectors of equal length.
    Parameters
    ----------
    x,y
        Arrays to calculate the nmrse between.
    Returns
    -------
    float
        Normalized root mean squared error.
    """

    max_val = np.max((np.max(x, axis=0), np.max(y, axis=0)))
    min_val = np.min((np.min(x, axis=0), np.min(y, axis=0)))

    diff = x - y

    return np.sqrt(np.sum(diff ** 2, axis=0)) / (max_val - min_val)


#########
# Tests #
#########


def test_2_1_operator():
    """Testing operator functionality of compute graph class:

    See Also
    --------
    :method:`add_operator`: Detailed documentation of method for adding operations to instance of `ComputeGraph`.
    """

    # simulation parameters
    dt = 1e-1
    sim_time = 10.0
    sim_steps = int(sim_time / dt)
    inp = np.zeros((sim_steps,)) + 0.5

    for b in backends:

        # test correct numerical evaluation of operator with two coupled simple, linear equations
        #########################################################################################

        # simulate operator behavior
        results = integrate("model_templates.test_resources.test_backend.net0", simulation_time=sim_time, step_size=dt,
                            outputs={'a': 'pop0/op0/a'}, vectorize=False, backend=b, clear=True, file_name='net0')

        # generate target values
        update0_1 = lambda x: x * 0.5
        update0_0 = lambda x: x + 2.0
        targets = np.zeros((sim_steps, 2), dtype=np.float64)
        for i in range(sim_steps-1):
            targets[i + 1, 0] = targets[i, 0] + dt * update0_0(targets[i, 1])
            targets[i + 1, 1] = targets[i, 1] + dt * update0_1(targets[i, 0])

        diff = results['a'].values[:] - targets[:, 1]
        assert np.mean(np.abs(diff)) == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of operator with a single differential equation and external input
        ######################################################################################################

        # simulate operator behavior
        results = integrate("model_templates.test_resources.test_backend.net1", simulation_time=sim_time, step_size=dt,
                            inputs={'pop0/op1/u': inp}, outputs={'a': 'pop0/op1/a'}, vectorize=False, backend=b,
                            clear=True, file_name='net1')

        # calculate operator behavior from hand
        update1 = lambda x, y: x + dt * (y - x)
        targets = np.zeros((sim_steps, 1), dtype=np.float32)
        for i in range(sim_steps-1):
            targets[i + 1] = update1(targets[i], inp[i])

        diff = results['a'].values[:] - targets[:, 0]
        assert np.mean(np.abs(diff)) == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of operator with two coupled equations (1 ODE, 1 non-DE eq.)
        ################################################################################################

        results = integrate("model_templates.test_resources.test_backend.net2", simulation_time=sim_time,
                            outputs={'a': 'pop0/op2/a'}, step_size=dt, vectorize=False, backend=b,
                            clear=True, file_name='net_2')

        # calculate operator behavior from hand
        update2 = lambda x: 1. / (1. + np.exp(-x))
        targets = np.zeros((sim_steps, 2), dtype=np.float32)
        for i in range(sim_steps-1):
            targets[i + 1, 1] = update2(targets[i, 0])
            targets[i + 1, 0] = update1(targets[i, 0], targets[i + 1, 1])

        diff = results['a'].values[:] - targets[:, 0]
        assert np.mean(np.abs(diff)) == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of operator with a two coupled DEs
        ######################################################################

        results = integrate("model_templates.test_resources.test_backend.net3", simulation_time=sim_time,
                            outputs={'b': 'pop0/op3/b'}, inputs={'pop0/op3/u': inp}, out_dir="/tmp/log",
                            step_size=dt, vectorize=True, backend=b, clear=True, file_name='net_3')

        # calculate operator behavior from hand
        update3_0 = lambda a, b, u: a + dt * (-10. * a + b ** 2 + u)
        update3_1 = lambda b, a: b + dt * 0.1 * a
        targets = np.zeros((sim_steps, 2), dtype=np.float32)
        for i in range(sim_steps-1):
            targets[i + 1, 0] = update3_0(targets[i, 0], targets[i, 1], inp[i])
            targets[i + 1, 1] = update3_1(targets[i, 1], targets[i, 0])

        diff = results['b'].values[:] - targets[:, 1]
        assert np.mean(np.abs(diff)) == pytest.approx(0., rel=accuracy, abs=accuracy)


def test_2_2_node():
    """Testing node functionality of compute graph class.

    See Also
    --------
    :method:`add_node`: Detailed documentation of method for adding nodes to instance of `ComputeGraph`.
    """

    dt = 1e-1
    sim_time = 10.
    sim_steps = int(np.round(sim_time/dt))

    for b in backends:

        # test correct numerical evaluation of node with 2 operators, where op1 projects to op2
        #######################################################################################

        # simulate node behavior
        results = integrate("model_templates.test_resources.test_backend.net4", simulation_time=sim_time,
                            outputs={'a': 'pop0/op1/a'}, step_size=dt, vectorize=True, backend=b, clear=True,
                            file_name='net4')

        # calculate node behavior from hand
        update0 = lambda x: x + dt * 2.
        update1 = lambda x, y: x + dt * (y - x)
        targets = np.zeros((sim_steps, 2), dtype=np.float32)
        for i in range(sim_steps-1):
            targets[i + 1, 0] = update0(targets[i, 0])
            targets[i + 1, 1] = update1(targets[i, 1], targets[i, 0])

        diff = results['a'].values[:] - targets[:, 1]
        assert np.mean(np.abs(diff)) == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of node with 2 independent operators
        ########################################################################

        # simulate node behavior
        results = integrate("model_templates.test_resources.test_backend.net5", simulation_time=sim_time,
                            outputs={'a': 'pop0/op5/a'}, step_size=dt, vectorize=True, backend=b, clear=True,
                            file_name='net5')

        # calculate node behavior from hand
        targets = np.zeros((sim_steps, 2), dtype=np.float32)
        for i in range(sim_steps-1):
            targets[i + 1, 0] = update0(targets[i, 0])
            targets[i + 1, 1] = update1(targets[i, 1], 0.)

        diff = results['a'].values[:] - targets[:, 1]
        assert np.mean(np.abs(diff)) == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of node with 2 independent operators projecting to the same target operator
        ###############################################################################################################

        results = integrate("model_templates.test_resources.test_backend.net6", simulation_time=sim_time,
                            outputs={'a': 'pop0/op1/a'}, step_size=dt, vectorize=True, backend=b, clear=True,
                            file_name='net6')

        # calculate node behavior from hand
        targets = np.zeros((sim_steps, 3), dtype=np.float32)
        update2 = lambda x: x + dt * (4. + np.tanh(0.5))
        for i in range(sim_steps-1):
            targets[i + 1, 0] = update0(targets[i, 0])
            targets[i + 1, 1] = update2(targets[i, 1])
            targets[i + 1, 2] = update1(targets[i, 2], targets[i, 0] + targets[i, 1])

        diff = results['a'].values[:] - targets[:, 2]
        assert np.mean(np.abs(diff)) == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of node with 1 source operator projecting to 2 independent targets
        ######################################################################################################

        results = integrate("model_templates.test_resources.test_backend.net7", simulation_time=sim_time,
                            outputs={'a': 'pop0/op1/a', 'b': 'pop0/op3/b'}, step_size=dt, vectorize=True,
                            backend=b, clear=True, file_name='net7')

        # calculate node behavior from hand
        targets = np.zeros((sim_steps, 4), dtype=np.float32)
        update3 = lambda a, b, u: a + dt * (-10. * a + b ** 2 + u)
        update4 = lambda x, y: x + dt * 0.1 * y
        for i in range(sim_steps-1):
            targets[i + 1, 0] = update0(targets[i, 0])
            targets[i + 1, 1] = update1(targets[i, 1], targets[i, 0])
            targets[i + 1, 2] = update3(targets[i, 2], targets[i, 3], targets[i, 0])
            targets[i + 1, 3] = update4(targets[i, 3], targets[i, 2])

        diff = np.mean(np.abs(results['a'].values[:] - targets[:, 1])) + \
               np.mean(np.abs(results['b'].values[:] - targets[:, 3]))
        assert diff == pytest.approx(0., rel=accuracy, abs=accuracy)


def test_2_3_edge():
    """Testing edge functionality of compute graph class.

    See Also
    --------
    :method:`add_edge`: Detailed documentation of add_edge method of `ComputeGraph`class.

    """

    dt = 1e-1
    sim_time = 10.
    sim_steps = int(np.round(sim_time/dt))
    inp = np.zeros((sim_steps, 1)) + 0.5

    final_results_comparison = []
    for b in backends:

        # test correct numerical evaluation of graph with 1 source projecting unidirectional to 2 target nodes
        ######################################################################################################

        # calculate edge behavior from hand
        update0 = lambda x, y: x + dt * y * 0.5
        update1 = lambda x, y: x + dt * (y + 2.0)
        update2 = lambda x, y: x + dt * (y - x)
        targets = np.zeros((sim_steps, 4), dtype=np.float32)
        for i in range(sim_steps-1):
            targets[i + 1, 0] = update0(targets[i, 0], targets[i, 1])
            targets[i + 1, 1] = update1(targets[i, 1], targets[i, 0])
            targets[i + 1, 2] = update2(targets[i, 2], targets[i, 0] * 2.0)
            targets[i + 1, 3] = update2(targets[i, 3], targets[i, 0] * 0.5)

        # simulate edge behavior
        results = integrate("model_templates.test_resources.test_backend.net8", simulation_time=sim_time,
                            outputs={'a': 'pop1/op1/a', 'b': 'pop2/op1/a'}, step_size=dt, vectorize=False,
                            backend=b, clear=True, file_name='net8')

        diff = np.mean(np.abs(results['a'].values[:] - targets[:, 2])) + \
               np.mean(np.abs(results['b'].values[:] - targets[:, 3]))
        assert diff == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of graph with 2 bidirectionaly coupled nodes
        ################################################################################

        results = integrate("model_templates.test_resources.test_backend.net9", simulation_time=sim_time,
                            outputs={'a': 'pop0/op1/a', 'b': 'pop1/op7/a'}, inputs={'pop1/op7/inp': inp},
                            step_size=dt, vectorize=False, backend=b, clear=True, file_name='net9')

        # calculate edge behavior from hand
        update3 = lambda x, y, z: x + dt * (y + z - x)
        targets = np.zeros((sim_steps, 2), dtype=np.float32)
        for i in range(sim_steps-1):
            targets[i + 1, 0] = update2(targets[i, 0], targets[i, 1] * 0.5)
            targets[i + 1, 1] = update3(targets[i, 1], targets[i, 0] * 2.0, inp[i])

        diff = np.mean(np.abs(results['a'].values[:] - targets[:, 0])) + \
               np.mean(np.abs(results['b'].values[:] - targets[:, 1]))
        assert diff == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of graph with 2 bidirectionally delay-coupled nodes
        #######################################################################################

        results = integrate("model_templates.test_resources.test_backend.net10", simulation_time=sim_time,
                            outputs={'a': 'pop0/op8/a', 'b': 'pop1/op8/a'}, step_size=dt, vectorize=False,
                            backend=b, clear=True, file_name='net10')

        # calculate edge behavior from hand
        delay0 = int(0.5 / dt)
        delay1 = int(1. / dt)
        targets = np.zeros((sim_steps, 2), dtype=np.float32)
        update4 = lambda x, y: x + dt * (2.0 + y)
        for i in range(sim_steps-1):
            inp0 = 0. if i < delay0 else targets[i - delay0, 1]
            inp1 = 0. if i < delay1 else targets[i - delay1, 0]
            targets[i + 1, 0] = update4(targets[i, 0], inp0 * 0.5)
            targets[i + 1, 1] = update4(targets[i, 1], inp1 * 2.0)

        diff = np.mean(np.abs(results['a'].values[:] - targets[:, 0])) + \
               np.mean(np.abs(results['b'].values[:] - targets[:, 1]))
        assert diff == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of graph with delay distributions
        #####################################################################

        results = integrate("model_templates.test_resources.test_backend.net13", simulation_time=sim_time,
                            outputs={'a1': 'p1/op9/a', 'a2': 'p2/op10/a'}, inputs={'p1/op9/I_ext': inp},
                            vectorize=False, step_size=dt, backend=b, solver='euler', clear=True,
                            file_name='net11')
        final_results_comparison.append(results.values)

    if len(final_results_comparison) > 1:
        r0 = final_results_comparison[0]
        final_comparison = np.mean([r0 - r for r in final_results_comparison[1:]])
        assert final_comparison == pytest.approx(0.0, rel=accuracy, abs=accuracy)


def test_2_4_solver():
    """Testing different numerical solvers of pyrates.

    See Also
    --------
    :method:`_solve`: Detailed documentation of how to numerical integration is performed by the `NumpyBackend`.
    :method:`run`: Detailed documentation of the method that needs to be called to solve differential equations in the
    `NumpyBackend`.
    """

    # define input
    dt = 1e-4
    dts = 1e-1
    sim_time = 20.
    sim_steps = int(np.round(sim_time / dt, decimals=0))
    inp = np.zeros((sim_steps, 1)) + 0.5

    for b in backends:

        # standard euler solver (trusted)
        r = integrate("model_templates.test_resources.test_backend.net13", simulation_time=sim_time,
                      outputs={'a1': 'p1/op9/a', 'a2': 'p2/op10/a'}, inputs={'p1/op9/I_ext': inp},
                      vectorize=False, step_size=dt, backend=b, solver='euler', clear=True, file_name='euler_solver',
                      sampling_step_size=dts)

        # scipy solver (tested)
        r2 = integrate("model_templates.test_resources.test_backend.net13", simulation_time=sim_time,
                       outputs={'a1': 'p1/op9/a', 'a2': 'p2/op10/a'}, inputs={'p1/op9/I_ext': inp}, method='RK23',
                       vectorize=False, step_size=dt, backend=b, solver='scipy', clear=True, file_name='scipy_solver',
                       sampling_step_size=dts)

        # Heun's method (tested)
        r3 = integrate("model_templates.test_resources.test_backend.net13", simulation_time=sim_time,
                       outputs={'a1': 'p1/op9/a', 'a2': 'p2/op10/a'}, inputs={'p1/op9/I_ext': inp},
                       vectorize=False, step_size=dt, backend=b, solver='heun', clear=True, file_name='heun_solver',
                       sampling_step_size=dts)

        assert np.mean(r.loc[:, 'a2'].values - r2.loc[:, 'a2'].values) == pytest.approx(0., rel=accuracy, abs=accuracy)
        assert np.mean(r.loc[:, 'a2'].values - r3.loc[:, 'a2'].values) == pytest.approx(0., rel=accuracy, abs=accuracy)


def test_2_5_inputs_outputs():
    """Tests the input-output interface of the run method in circuits of different hierarchical depth.

    See Also
    -------
    :method:`CircuitIR.run` detailed documentation of how to use the arguments `inputs` and `outputs`.

    """

    dt = 1e-3
    dts = 1e-1
    sim_time = 10.
    sim_steps = int(np.round(sim_time / dt, decimals=0))
    inp = np.zeros((sim_steps, 1)) + 0.5

    for b in backends:

        # define inputs and outputs for each population separately
        ##########################################################

        # perform simulation
        r1 = integrate("model_templates.test_resources.test_backend.net13", simulation_time=sim_time,
                       outputs={'a1': 'p1/op9/a'}, inputs={'p1/op9/I_ext': inp}, vectorize=True, step_size=dt,
                       backend=b, solver='euler', clear=True, file_name='inout_1', sampling_step_size=dts)

        # define input and output for both populations simultaneously
        #############################################################

        # perform simulation
        r2 = integrate("model_templates.test_resources.test_backend.net13", simulation_time=sim_time,
                       outputs=['all/op9/a'], inputs={'all/op9/I_ext': inp}, vectorize=True, step_size=dt, backend=b,
                       solver='euler', clear=True, file_name='inout_2', sampling_step_size=dts)

        assert np.mean(r1.values.flatten() - r2.values.flatten()) == pytest.approx(0., rel=accuracy, abs=accuracy)

        # repeat in a network with 2 hierarchical levels of node organization
        #####################################################################

        # define input
        inp2 = np.zeros((sim_steps, 1)) + 0.1

        # perform simulation
        r1 = integrate("model_templates.test_resources.test_backend.net14", simulation_time=sim_time, vectorize=True,
                       step_size=dt, backend=b, solver='euler', clear=True, sampling_step_size=dts,
                       outputs={'a1': 'c1/p1/op9/a', 'a2': 'c1/p2/op10/a', 'a3': 'c2/p1/op9/a', 'a4': 'c2/p2/op10/a'},
                       inputs={'c1/p1/op9/I_ext': inp, 'c1/p2/op10/I_ext': inp2, 'c2/p1/op9/I_ext': inp,
                               'c2/p2/op10/I_ext': inp2}, file_name='inout_3')

        # perform simulation
        r2 = integrate("model_templates.test_resources.test_backend.net14", simulation_time=sim_time,
                       outputs={'a1': 'all/all/op9/a', 'a2': 'all/all/op10/a'},
                       inputs={'all/all/op9/I_ext': inp, 'all/all/op10/I_ext': inp2},
                       vectorize=True, step_size=dt, backend=b, solver='euler', clear=True, file_name='inout_4',
                       sampling_step_size=dts)

        assert np.mean(r1.values.flatten() - r2.values.flatten()) == pytest.approx(0., rel=accuracy, abs=accuracy)


def test_2_6_vectorization():
    """Tests whether a Jansen-Rit-based circuit with and without vectorization of mathematical operations yields
    identical results.

    See Also
    --------
    :method:`CircuitTemplate.run` for a documentation of the keyword argument `vectorize`.
    """

    dt = 1e-4
    dts = 1e-2
    T = 1.0
    inp = np.zeros((int(np.round(T/dt)),)) + 220.0

    for i, b in enumerate(backends):

        # simulation without vectorization of the network equations
        r1 = integrate("model_templates.neural_mass_models.jansenrit.JRC_2delaycoupled", vectorize=False,
                       inputs={"jrc2/pc/rpo_e_in/u": inp}, outputs={"r": "jrc1/ein/rpo_e/v"}, backend=b,
                       solver='euler', step_size=dt, clear=True, simulation_time=T, sampling_step_size=dts,
                       file_name=f'novec{i + 1}')

        # simulation with vectorized network equations
        r2 = integrate("model_templates.neural_mass_models.jansenrit.JRC_2delaycoupled", vectorize=True,
                       inputs={"jrc2/pc/rpo_e_in/u": inp}, outputs={"r": "jrc1/ein/rpo_e/v"}, backend=b,
                       solver='euler', step_size=dt, clear=True, simulation_time=T, sampling_step_size=dts,
                       file_name=f'vec{i + 1}')

        assert np.mean(r1.values - r2.values) == pytest.approx(0.0, rel=accuracy, abs=accuracy)


def test_2_7_backends():
    """Tests the whether different backends produce comparable results when simulating the dynamics of different models.

    See Also
    -------
    :method:`CircuitIR.__init__` for documentation of the available backend options.
    """

    dt = 5e-4
    dts = 1e-3
    T = 10.

    r0 = integrate("model_templates.neural_mass_models.qif.qif_sfa", simulation_time=T, sampling_step_size=dts,
                   inputs=None, outputs={"r": "p/qif_sfa_op/r"}, solver='euler', step_size=dt, clear=True,
                   file_name='m0', vectorize=False)

    for i, b in enumerate(backends):

        if b != 'default':

            r = integrate("model_templates.neural_mass_models.qif.qif_sfa",
                          inputs=None, outputs={"r": "p/qif_sfa_op/r"}, backend=b, solver='euler', step_size=dt,
                          clear=True, simulation_time=T, sampling_step_size=dts, file_name=f'm{i+1}', vectorize=False)

            assert np.mean(r0.values - r.values) == pytest.approx(0.0, rel=accuracy, abs=accuracy)



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

"""Wraps Julia such that its low-level functions can be used by PyRates to create and simulate a compute graph.
"""

# pyrates internal _imports
import sys

from ..base import BaseBackend
from ..computegraph import ComputeVar
from .julia_funcs import julia_funcs

# external _imports
from typing import Optional, Dict, List, Callable, Iterable, Union, Tuple
import numpy as np

# meta infos
__author__ = "Richard Gast"
__status__ = "development"


# backend classes
#################


class JuliaBackend(BaseBackend):

    def __init__(self,
                 ops: Optional[Dict[str, str]] = None,
                 imports: Optional[List[str]] = None,
                 **kwargs
                 ) -> None:
        """Instantiates Julia backend.
        """

        # add user-provided operations to function dict
        julia_ops = julia_funcs.copy()
        if ops:
            julia_ops.update(ops)

        # set default float precision to float64
        kwargs["float_precision"] = "float64"

        # call parent method
        super().__init__(ops=julia_ops, imports=imports, file_ending='.jl', start_idx=1, add_hist_arg=True, **kwargs)

        # define julia-specific imports
        self._imports.pop(0)
        self._imports.append("using LinearAlgebra")

        # set up pyjulia
        from julia.api import Julia
        jl = Julia(runtime=kwargs.pop('julia_path'), compiled_modules=False)
        from julia import Main
        self._jl = Main
        self._no_vectorization = ["*(", "interp("]
        self._fcall = None

    def get_var(self, v: ComputeVar):
        v = super().get_var(v)
        dtype = v.dtype.name
        s = sum(v.shape)
        if s > 0:
            return v
        if 'float' in dtype:
            return float(v)
        if 'complex' in dtype:
            return complex(np.real(v), np.imag(v))
        return int(v)

    def add_var_update(self, lhs: ComputeVar, rhs: str, lhs_idx: Optional[str] = None, rhs_shape: Optional[tuple] = ()):

        super().add_var_update(lhs=lhs, rhs=rhs, lhs_idx=lhs_idx, rhs_shape=rhs_shape)
        if rhs_shape or lhs_idx:
            line = self.code.pop()
            lhs, rhs = line.split(' = ')
            if not any([rhs[:len(expr)] == expr for expr in self._no_vectorization]):
                rhs = f"@. {rhs}"
            self.add_code_line(f"{lhs} = {rhs}")

    def add_var_hist(self, lhs: str, delay: Union[ComputeVar, float], state_idx: Union[int, tuple], **kwargs):
        idx = self._process_idx(state_idx)
        d = self._process_delay(delay)
        self.add_code_line(f"{lhs} = hist((), t-{d}; idxs={idx})")

    def get_hist_func(self, y: np.ndarray):
        self._jl.eval(f"y_init = {y.tolist()}")
        hist = """
        function hist(p, t; idxs=nothing)
            return idxs == nothing ? y_init : y_init[idxs]
        end
        """
        self._jl.eval(hist)
        return self._jl.hist

    def create_index_str(self, idx: Union[str, int, tuple], separator: str = ',', apply: bool = True,
                         **kwargs) -> Tuple[str, dict]:

        if not apply:
            self._start_idx = 0
            idx, idx_dict = super().create_index_str(idx, separator, apply, **kwargs)
            self._start_idx = 1
            return idx, idx_dict
        else:
            return super().create_index_str(idx, separator, apply, **kwargs)

    def generate_func_tail(self, rhs_var: str = 'dy'):

        self.add_code_line(f"return {rhs_var}")
        self.remove_indent()
        self.add_code_line("end")

    def generate_func(self, func_name: str, to_file: bool = True, **kwargs):

        self._fcall = func_name

        # generate the current function string via the code generator
        if kwargs.pop('julia_ode', False):
            self.add_linebreak()
            self.add_code_line(f"function {func_name}_julia(dy, y, p, t)")
            self.add_indent()
            self.add_code_line(f"return {func_name}(t, y, dy, p...)")
            self.remove_indent()
            self.add_code_line("end")
        if kwargs.pop('julia_dde', False):
            self.add_linebreak()
            self.add_code_line(f"function {func_name}_julia(dy, y, h, p, t)")
            self.add_indent()
            self.add_code_line(f"return {func_name}(t, y, h, dy, p...)")
            self.remove_indent()
            self.add_code_line("end")
        func_str = self.generate()

        if to_file:

            # save rhs function to file
            file = f'{self.fdir}/{self._fname}{self._fend}' if self.fdir else f"{self._fname}{self._fend}"
            with open(file, 'w') as f:
                f.writelines(func_str)
                f.close()

            # import function from file
            rhs_eval = self._jl.include(file)

        else:

            # just execute the function string, without writing it to file
            rhs_eval = self._jl.eval(func_str)

        # apply function decorator
        decorator = kwargs.pop('decorator', None)
        if decorator:
            decorator_kwargs = kwargs.pop('decorator_kwargs', dict())
            rhs_eval = decorator(rhs_eval, **decorator_kwargs)

        return rhs_eval

    def _solve(self, solver: str, func: Callable, args: tuple, T: float, dt: float, dts: float, y0: np.ndarray,
               t0: np.ndarray, times: np.ndarray, **kwargs) -> np.ndarray:

        if 'julia' in solver:

            # solve via DifferentialEquations.jl
            self._jl.eval('using DifferentialEquations')

            if 'dde' in solver:

                # define wrapper function and solver family
                jfunc = f"""
                function julia_dderun(du,u,h,p,t)
                    return {self._fcall}(t,u,h,du,p...)
                end
                """

                # solve ivp via DifferentialEquations.jl solver
                self._jl.eval(jfunc)
                model = self._jl.DDEProblem(self._jl.julia_dderun, y0, args[0], [0.0, T], args[2:])
                method = kwargs.pop('method', 'Tsit5')
                solver = getattr(self._jl, method)
                solver = self._jl.MethodOfSteps(solver())
                atol, rtol = kwargs.pop('atol', 1e-6), kwargs.pop('rtol', 1e-3)
                results = self._jl.solve(model, solver, saveat=times, atol=atol, rtol=rtol)

            else:

                # define wrapper function and solver family
                jfunc = f"""
                function julia_oderun(du,u,p,t)
                    return {self._fcall}(t,u,du,p...)
                end
                """

                # solve ivp via DifferentialEquations.jl solver
                self._jl.eval(jfunc)
                model = self._jl.ODEProblem(self._jl.julia_oderun, y0, [0.0, T], args[1:])
                method = kwargs.pop('method', 'Tsit5')
                solver = getattr(self._jl, method)
                atol, rtol = kwargs.pop('atol', 1e-6), kwargs.pop('rtol', 1e-3)
                results = self._jl.solve(model, solver(), saveat=times, atol=atol, rtol=rtol)

            results = np.asarray(results).T

        else:

            # non-julia solver
            results = super()._solve(solver=solver, func=func, args=args, T=T, dt=dt, dts=dts, y0=y0, t0=t0,
                                     times=times, **kwargs)

        return results

    def _add_func_call(self, name: str, args: Iterable, return_var: str = 'dy'):
        self.add_code_line(f"function {name}({','.join(args)})")

    def _process_idx(self, idx: Union[Tuple[int, int], int, str, ComputeVar], **kwargs) -> str:

        if type(idx) is str and idx != ':' and ':' in idx:
            idx0, idx1 = idx.split(':')
            self._start_idx = 0
            idx0 = int(self._process_idx(idx0))
            idx1 = int(self._process_idx(idx1))
            self._start_idx = 1
            return self._process_idx((idx0, idx1))
        if type(idx) is ComputeVar and idx.name == "t" and idx.value >= self._start_idx:
            self._start_idx = 0
            idx_processed = super()._process_idx(idx=idx, **kwargs)
            self._start_idx = 1
            return idx_processed
        return super()._process_idx(idx=idx, **kwargs)

    @staticmethod
    def expr_to_str(expr: str, args: tuple):

        # replace power operator
        func = '**'
        while func in expr:
            expr = expr.replace(func, '^')

        return expr



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation
"""Functions for performing parameter grid simulations with pyrates model_templates.
"""

# meta infos
__author__ = "Christoph Salomon, Richard Gast"
__status__ = "development"

# external imports
from typing import Union, Optional
import pandas as pd
import numpy as np
from copy import deepcopy

# pyrates imports
from pyrates.frontend import CircuitTemplate, OperatorTemplate, template
from pyrates.ir import clear_ir_caches


#####################################
# temporary file clearing functions #
#####################################


def clear(model: CircuitTemplate, **kwargs):
    """Function that clears all temporary files and caches that have been created via a PyRates model.

    Parameters
    ----------
    model
        Instance of a `CircuitTemplate`.
    kwargs
        Additional keyword arguments to be passed to `clear_frontend_caches`.
    """
    try:
        model.clear()
    except AttributeError:
        pass
    clear_frontend_caches(**kwargs)


def clear_frontend_caches(clear_template_cache=True, clear_ir_cache=True):
    """Utility to clear caches in the frontend.

    Parameters
    ----------
    clear_template_cache
        toggles whether or not to clear the template_cache that contains all previously loaded templates
    clear_ir_cache
        toggles whether or not to clear the cache of unique OperatorIR instances
    """
    if clear_template_cache:
        template.clear_cache()

    if clear_ir_cache:
        OperatorTemplate.cache.clear()
        clear_ir_caches()


########################
# simulation functions #
########################


def integrate(circuit: Union[str, CircuitTemplate], **kwargs):
    """Directly simulate dynamics of a circuit."""
    if type(circuit) is str:
        circuit = CircuitTemplate.from_yaml(circuit)
    results = circuit.run(**kwargs)
    if 'clear' in kwargs and kwargs['clear']:
        clear_frontend_caches()
    return results


#############################
# parameter sweep functions #
#############################

def linearize_grid(grid: dict, permute: bool = False) -> pd.DataFrame:
    """Turns the grid into a grid that can be traversed linearly, i.e. pairwise.

    Parameters
    ----------
    grid
        Parameter grid.
    permute
        If true, all combinations of the parameter values in grid will be created.

    Returns
    -------
    pd.DataFrame
        Resulting linear grid in form of a data frame.

    """

    arg_lengths = [len(arg) for arg in grid.values()]

    if len(list(set(arg_lengths))) == 1 and not permute:
        return pd.DataFrame(grid)
    elif permute:
        vals, keys = [], []
        for key, val in grid.items():
            vals.append(val)
            keys.append(key)
        new_grid = np.stack(np.meshgrid(*tuple(vals)), -1).reshape(-1, len(grid))
        return pd.DataFrame(new_grid, columns=keys)
    else:
        raise ValueError('Wrong number of parameter combinations. If `permute` is False, all parameter vectors in grid '
                         'must have the same number of elements.')


def adapt_circuit(circuit: Union[CircuitTemplate, str], params: dict, param_map: dict) -> CircuitTemplate:
    """Changes the parametrization of a circuit.

    Parameters
    ----------
    circuit
        Circuit instance.
    params
        Key-value pairs of the parameters that should be changed.
    param_map
        Map between the keys in params and the circuit variables.

    Returns
    -------
    CircuitIR
        Updated circuit instance.

    """

    if type(circuit) is str:
        circuit = deepcopy(CircuitTemplate.from_yaml(circuit))
    else:
        circuit = deepcopy(circuit)

    node_updates = {}
    edge_updates = []

    for key in params.keys():

        val = params[key]

        if 'nodes' in param_map[key]:
            for n in param_map[key]['nodes']:
                for v in param_map[key]['vars']:
                    node_updates[f"{n}/{v}"] = val
        else:
            edges = param_map[key]['edges']
            if len(edges[0]) < 3:
                for source, target in edges:
                    for var in param_map[key]['vars']:
                        edge = circuit.get_edge(source=source, target=target, idx=0)
                        edge_updates.append((edge[0], edge[1], {var: val}))
            else:
                for source, target, idx in edges:
                    for var in param_map[key]['vars']:
                        edge = circuit.get_edge(source=source, target=target, idx=idx)
                        edge_updates.append((edge[0], edge[1], {var: val}))

    return circuit.update_var(node_vars=node_updates, edge_vars=edge_updates)


def grid_search(circuit_template: Union[CircuitTemplate, str], param_grid: Union[dict, pd.DataFrame], param_map: dict,
                step_size: float, simulation_time: float, outputs: dict, inputs: Optional[dict] = None,
                sampling_step_size: Optional[float] = None, permute_grid: bool = False, **kwargs) -> tuple:
    """Function that runs multiple parametrizations of the same circuit in parallel and returns a combined output.

    Parameters
    ----------
    circuit_template
        Path to the circuit template.
    param_grid
        Key-value pairs for each circuit parameter that should be altered over different circuit parametrizations.
    param_map
        Key-value pairs that map the keys of param_grid to concrete circuit variables.
    step_size
        Simulation step-size in s.
    simulation_time
        Simulation time in s.
    outputs
        Output variables as provided to the `run` method of `:class:ComputeGraph`.
    inputs
        Extrinsic inputs as provided to the `run` method of `:class:ComputeGraph`.
    sampling_step_size
        Sampling step-size as provided to the `run` method of `:class:ComputeGraph`.
    permute_grid
        If true, all combinations of the provided param_grid values will be realized. If false, the param_grid values
        will be traversed pairwise.
    kwargs
        Additional keyword arguments passed to the `CircuitTemplate.run` call.


    Returns
    -------
    tuple
        Simulation results stored in a multi-index data frame, and the mapping between the data frame column names and
        the parameter grid.
    """

    # argument pre-processing
    #########################

    vectorize = kwargs.pop('vectorize', True)

    # linearize parameter grid if necessary
    if type(param_grid) is dict:
        param_grid = linearize_grid(param_grid, permute_grid)

    # create grid-structure of network
    ##################################

    # get parameter names and grid length
    param_keys = list(param_grid.keys())

    # assign parameter updates to each circuit, combine them to unconnected network and remember their parameters
    circuit_names = []
    circuit = CircuitTemplate(name='top_lvl', path='none')
    for idx in param_grid.index:
        new_params = {}
        for key in param_keys:
            new_params[key] = param_grid[key][idx]
        circuit_tmp = adapt_circuit(circuit_template, new_params, param_map)
        circuit_key = f'{circuit_tmp.name}_{idx}'
        circuit = circuit.update_template(circuits={circuit_key: circuit_tmp})
        circuit_names.append(circuit_key)
    param_grid.index = circuit_names

    # adjust input of simulation to combined network
    if inputs:
        for inp_key, inp in inputs.copy().items():
            inputs[f"all/{inp_key}"] = inp
            inputs.pop(inp_key)

    # adjust output of simulation to combined network
    outputs_new = {}
    for key, out in outputs.items():
        outputs_new[key] = f"all/{out}"

    # simulate the circuits behavior
    results = circuit.run(simulation_time=simulation_time,
                          step_size=step_size,
                          sampling_step_size=sampling_step_size,
                          inputs=inputs,
                          outputs=outputs_new,
                          vectorize=vectorize,
                          **kwargs)    # type: pd.DataFrame

    # return results
    return results, param_grid


class Interactive2DParamPlot(object):
    def __init__(self, data_map: np.array, data_series: pd.DataFrame, x_values: np.array, y_values: np.array,
                 x_key: str, y_key: str, param_map: pd.DataFrame, tmin=0., title=None, **kwargs):
        """Creates an interactive 2D plot that allows visualization of time series using button press events

        Derive child class and change get_data() respectively to utilize this plotting method

        Parameters
        ----------
        data_map
            2D ndarray containing a value based on each column data_series, respectively.
        data_series
            DataFrame containing all data series used to create the data map
        x_values
            ndarray containing values of data-map columns.
        y_values
            ndarray containing values of data-map rows.
        x_key
        y_key
        param_map
            Dataframe containing the mapping between data-series columns (index) and x/y value pairs (columns)
            (as returned by `grid_search`).
        tmin
            Starting point for time-series plots in time units (float).
        title
            Title of 2D plot.
        kwargs
            Additional information to access a column in data_series if necessary

        Returns
        -------

        """

        import matplotlib.pyplot as plt

        dt = kwargs.pop('step_size', data_series.index[1] - data_series.index[0])
        state_var = kwargs.pop('state_var', '')
        tmin = int(tmin/dt)
        self.data = data_series.iloc[tmin:, :]
        self.x_values = x_values
        self.y_values = y_values

        # set up param map matrix
        self.map = pd.DataFrame(columns=self.x_values, index=self.y_values)
        for key in param_map.index:
            x, y = param_map.loc[key, x_key], param_map.loc[key, y_key]
            x_idx = np.argmin(np.abs(self.x_values - x))
            y_idx = np.argmin(np.abs(self.y_values - y))
            self.map.iloc[y_idx, x_idx] = key

        # Create canvas
        if 'subplots' in kwargs:
            self.fig, self.ax = kwargs.pop('subplots')
        else:
            skwargs = {}
            for key in ["figsize", "gridspec"]:
                if key in kwargs:
                    skwargs[key] = kwargs.pop(key)
            self.fig, self.ax = plt.subplots(ncols=2, nrows=1, **skwargs)

        # plot signals into right subplot
        cmap = kwargs.pop("cmap_lines", "magma")
        self.line_colors = plt.get_cmap(cmap, lut=self.data[self.map.iloc[0, 0]].shape[-1]).colors
        self.update_lineplot(0, 0)

        # Initiate marker
        self.marker = self.ax[0].plot(0, 0, 'x', color='white', markersize='10')

        # Plot 2D data in left subplot
        num_x_ticks = kwargs.pop('num_x_ticks', len(x_values))
        num_y_ticks = kwargs.pop('num_y_ticks', len(y_values))
        shrink = kwargs.pop('cbar_shrink', 0.5)
        im = self.ax[0].imshow(data_map, **kwargs)
        set_num_axis_ticks(ax=self.ax[0], num_x_ticks=num_x_ticks, num_y_ticks=num_y_ticks)
        self.ax[0].set_xlabel(x_key)
        self.ax[0].set_ylabel(y_key)
        self.ax[0].set_xticklabels([""] + list(np.round(x_values, decimals=2)))
        self.ax[0].set_yticklabels([""] + list(np.round(y_values, decimals=2)))
        if title:
            self.ax[0].set_title(title)
        plt.colorbar(im, ax=self.ax[0], shrink=shrink)

        # Call Interactive2DPlot class instance when mouse button is pressed inside the 2D plot
        self.fig.canvas.mpl_connect('button_press_event', self)

    def __call__(self, event):
        """Try to access a column in data_series using x and y values based on cursor position

        Is called on mouse button press event. Converts the current cursor coordinates inside the plot into x and y
        values based on the data in x_values and y_values. x and y values are used to access a column in data_series.
        Access of data_series can be customized in self.get_data().

        :param event:
        :return:
        """

        # Only allow button press events in the (left) 2D overview plot
        if event.inaxes != self.ax[0]:
            return

        # Reset axes
        self.marker[0].remove()

        # Transform cursor coordinates in x and y values
        x_sample = event.xdata
        y_sample = event.ydata

        # Add marker at event coordinates
        self.marker = self.ax[0].plot(x_sample, y_sample, 'x', color='white', markersize='10')

        # Update serial plot
        self.update_lineplot(int(np.round(x_sample, decimals=0)), int(np.round(y_sample, decimals=0)))

        # redraw figure
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

    def update_lineplot(self, x, y):

        data = self.get_data(x, y)
        lines = self.ax[1].get_lines()
        if lines:
            for line, key in zip(self.ax[1].get_lines(), data.keys()):
                line.set_data(data.index, data[key].values)
        else:
            for i, key in enumerate(data.keys()):
                self.ax[1].plot(data.index, data[key].values, c=self.line_colors[i])
        self.ax[1].set_title(
            f'x: {np.round(self.x_values[x], decimals=2)}, y: {np.round(self.y_values[y], decimals=2)}')
        ymin, ymax = np.min(data.values), np.max(data.values)
        margin = (ymax - ymin)*0.02
        self.ax[1].set_ylim([ymin-margin, ymax+margin])

    def set_map_xlabel(self, label):
        self.ax[0].set_xlabel(label)

    def set_map_ylabel(self, label):
        self.ax[0].set_ylabel(label)

    def set_map_title(self, title):
        self.ax[0].set_title(title)

    def set_series_xlabel(self, label):
        self.ax[1].set_xlabel(label)

    def set_series_ylabel(self, label):
        self.ax[1].set_ylabel(label)

    def get_data(self, x, y):
        return self.data[self.map.iloc[y, x]]


def set_num_axis_ticks(ax, num_x_ticks=10, num_y_ticks=10):
    """Set the number of x and y ticks of a plot axis

    Parameters
    ----------
    ax
    num_x_ticks
    num_y_ticks

    Returns
    -------

    """
    from matplotlib.ticker import MaxNLocator
    ax.xaxis.set_major_locator(MaxNLocator(num_x_ticks))
    ax.yaxis.set_major_locator(MaxNLocator(num_y_ticks))

##########################
# file storage functions #
##########################


def create_directory(path):
    """check if a directory exists and create it otherwise"""

    import os
    import errno

    directory = os.path.dirname(path)
    if directory and not os.path.exists(directory):
        try:
            os.makedirs(directory)
        except OSError as exc:  # Guard against race condition
            if exc.errno != errno.EEXIST:
                raise



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

"""Wraps tensorflow such that it's low-level functions can be used by PyRates to create and simulate a compute graph.
"""

# pyrates internal _imports
from ..base import BaseBackend
from ..computegraph import ComputeVar
from .tensorflow_funcs import tf_funcs

# external _imports
from typing import Optional, Dict, Callable, List, Tuple, Union
import tensorflow as tf
import numpy as np

# meta infos
__author__ = "Richard Gast"
__status__ = "development"


#######################################
# classes for backend functionalities #
#######################################


class TensorflowBackend(BaseBackend):
    """Wrapper to tensorflow. This class provides an interface to all tensorflow functionalities that may be accessed
    via pyrates. All tensorflow variables and operations will be stored in a layered compute graph that can be executed
    to evaluate the (dynamic) behavior of the graph.

    Parameters
    ----------
    ops
        Additional operations this backend instance can perform, defined as key-value pairs. The key can then be used in
        every equation parsed into this backend. The value is a dictionary again, with two keys:
            1) name - the name of the function/operation (used during code generation)
            2) call - the call signature of the function including import abbreviations (i.e. `tf.add` for tensorflow's
            add function)
    dtypes
        Additional data-types this backend instance can use, defined as key-value pairs.
    name
        Name of the backend instance. Used during code generation to create a recognizable file structure.
    float_default_type
        Default float precision. If no data-type is indicated for a particular variable, this will be used.
    imports
        Can be used to pass additional import statements that are needed for code generation of the custom functions
        provided via `ops`. Will be added to the top of each generated code file.

    """

    def __init__(self,
                 ops: Optional[Dict[str, str]] = None,
                 imports: Optional[List[str]] = None,
                 **kwargs
                 ) -> None:
        """Instantiates tensorflow backend, i.e. a tensorflow graph.
        """

        # add user-provided operations to function dict
        tf_ops = tf_funcs.copy()
        if ops:
            tf_ops.update(ops)

        # call parent method
        super().__init__(ops=tf_ops, imports=imports, **kwargs)

    def get_var(self, v: ComputeVar):
        v_tmp = super().get_var(v)
        if v.vtype == 'constant':
            return tf.constant(name=v.name, value=v_tmp, dtype=v_tmp.dtype)
        else:
            return tf.Variable(name=v.name, initial_value=v_tmp, dtype=v_tmp.dtype)

    def add_var_update(self, lhs: ComputeVar, rhs: str, lhs_idx: Optional[str] = None, rhs_shape: Optional[tuple] = ()):

        lhs = lhs.name
        if lhs_idx:
            idx, _ = self.create_index_str(lhs_idx, scatter=True, size=rhs_shape[0] if rhs_shape else 1, apply=True)
            if not rhs_shape:
                rhs = f"[{rhs}]"
            else:
                try:
                    rhs = f"[{float(rhs)}]"
                except ValueError:
                    pass
            self.add_code_line(f"{lhs}.scatter_nd_update({idx}, {rhs})")
        else:
            self.add_code_line(f"{lhs} = {rhs}")

    def create_index_str(self, idx: Union[str, int, tuple], separator: str = ',', apply: bool = True,
                         scatter: bool = False, size: Optional[int] = 1, max_length: int = 2) -> Tuple[str, dict]:

        # case I: multiple indices
        if scatter:

            # recognize indices that already match tensorflow requirements
            if type(idx) is list:
                return f"{idx}", dict()

            # create tensorflow-specific index
            if type(idx) is not tuple:
                idx = tuple(idx.split(separator)) if separator in idx else (idx,)
            new_idx, return_dict = self._get_tf_idx(idx, size=size, max_length=max_length)
            return f"{new_idx}" if apply else new_idx, return_dict

        # case II: indexing via tf.gather_nd
        if type(idx) is tuple and any([type(i) is ComputeVar and i.is_constant for i in idx]):

            # create tensorflow-specific index
            new_idx, return_dict = self._get_tf_idx(idx, size=size, max_length=max_length)
            return f".gather_nd({new_idx})" if apply else new_idx, return_dict

        # case III: default indexing
        return super().create_index_str(idx=idx, separator=separator, apply=apply, scatter=False)

    def finalize_idx_str(self, var: ComputeVar, idx: str):
        gather = ".gather_nd("
        if gather in idx:
            start = idx.find(gather) + len(gather)
            stop = idx.find(")")
            idx_str = idx[start:stop]
            import_line = "from tensorflow import gather_nd"
            if import_line not in self._imports:
                self._imports.append(import_line)
            return f"gather_nd({var.name}, {idx_str})"
        return f"{var.name}{idx}"

    def _process_idx(self, idx: Union[Tuple[int, int], int, str, ComputeVar], size: int = 1, scatter: bool = False
                     ) -> Union[str, list]:

        # case I: tensorflow-specific mode where lists of indexing integers are collected for each variable axis
        if scatter:
            if type(idx) is ComputeVar:
                return (idx.value + self._start_idx).tolist()
            if type(idx) is tuple:
                idx_range = np.arange(idx[0] + self._start_idx, idx[1]).tolist()
                return idx_range
            if type(idx) is int:
                return [idx + self._start_idx]
            if idx == ':':
                return self._process_idx((self._start_idx, size), scatter=scatter)
            if ':' in idx:
                idx0, idx1 = idx.split(':')
                idx0 = self._process_idx(idx0, scatter=scatter)[0]
                idx1 = self._process_idx(idx1, scatter=scatter)[0]
                return self._process_idx((idx0, idx1), scatter=scatter)
            try:
                return self._process_idx(int(idx), scatter=scatter)
            except (ValueError, TypeError):
                return idx

        # case II: multiple indices into a single dimension
        if type(idx) is ComputeVar and idx.is_constant:
            return (idx.value + self._start_idx).tolist()

        # case III: default mode
        return super()._process_idx(idx)

    def _get_tf_idx(self, idx: tuple, size: int, max_length) -> Tuple[list, dict]:

        # collect axis indices
        idx = list(idx)
        for i in range(len(idx)):
            idx[i] = self._process_idx(idx[i], size=size, scatter=True)

        # bring indices into form required for `tf.Variable.scatter_nd_update`
        length = max([len(i) for i in idx])
        new_idx = []
        for i in range(length):
            idx_tmp = []
            for j in range(len(idx)):
                try:
                    idx_tmp.append(idx[j][i])
                except IndexError:
                    idx_tmp.append(idx[j][0])
            new_idx.append(idx_tmp)

        if len(new_idx) > max_length:

            # for many indices, create new indexing constant
            idx = self.idx_dummy_var
            return_dict = {idx: new_idx}

        else:

            return_dict = dict()

        return new_idx, return_dict

    def _solve_euler(self, func: Callable, args: tuple, T: float, dt: float, dts: float, y: np.ndarray, t0: np.ndarray):

        # preparations for fixed step-size integration
        zero = tf.constant(0, dtype=tf.int32)
        steps = tf.constant(int(np.round(T / dt)))
        store_steps = int(np.round(T / dts))
        store_step = tf.constant(int(np.round(dts / dt)))
        state_rec = tf.Variable(np.zeros((store_steps, y.shape[0])), dtype=y.dtype)
        dt = tf.constant(dt)

        return self._solve_euler_tf(func, args, steps, store_step, zero, state_rec, y, t0, dt).numpy()

    @staticmethod
    def _solve_scipy(func: Callable, args: tuple, T: float, dt: float, y: tf.Variable, t0: tf.Variable,
                     times: np.ndarray, **kwargs):

        # solve ivp via scipy methods (solvers of various orders with adaptive step-size)
        from scipy.integrate import solve_ivp
        kwargs['t_eval'] = times
        func = tf.function(func)
        dtype = y.dtype

        # wrapper function to scipy solver
        def f(t: float, y: np.ndarray):
            rhs = func(tf.constant(t, dtype=dtype), tf.constant(y, dtype=dtype), *args)
            return rhs.numpy()

        # call scipy solver
        results = solve_ivp(fun=f, t_span=(t0.numpy(), T), y0=y.numpy(), first_step=dt, **kwargs)
        return results['y'].T

    @staticmethod
    @tf.function
    def _solve_euler_tf(func: Callable, args: tuple, steps: tf.constant, store_step: int, zero: tf.constant,
                        state_rec: tf.Variable, y: tf.Variable, idx: tf.Variable, dt: tf.constant):

        for step in tf.range(steps):

            if tf.equal(tf.math.floormod(step, store_step), zero):
                state_rec.scatter_nd_update([[idx]], [y])
                idx.assign_add(1)

            rhs = func(step, y, *args)
            y.assign_add(dt*rhs)



"""Pytest module-wide configuration file."""

import pytest
import os


# @pytest.fixture(scope="session", autouse=True)
# def ensure_test_directory():
#     if os.getcwd().endswith("PyRates"):
#         os.chdir("tests/")


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test
    # Find and save current (test) working directory
    cwd = os.getcwd()
    # A test function will be run at this point
    yield
    # Code that will run after the test
    # revert changes to working directory
    os.chdir(cwd)



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

"""Wraps fortran such that it's low-level functions can be used by PyRates to create and simulate a compute graph.
"""

# pyrates internal _imports
from ..base import BaseBackend
from ..computegraph import ComputeVar
from .fortran_funcs import fortran_funcs
from ..parser import replace

# external _imports
import subprocess
import os
from numpy import f2py
import numpy as np
from typing import Optional, Dict, List, Union, Tuple, Iterable, Callable

# meta infos
__author__ = "Richard Gast"
__status__ = "development"


# backend classes
#################


class FortranBackend(BaseBackend):

    n1 = 62
    n2 = 72
    linebreak_start = "     & "
    linebreak_end = "&"

    def __init__(self,
                 ops: Optional[Dict[str, str]] = None,
                 imports: Optional[List[str]] = None,
                 **kwargs
                 ) -> None:
        """Instantiates Fortran backend.
        """

        # add user-provided operations to function dict
        fort_ops = fortran_funcs.copy()
        if ops:
            fort_ops.update(ops)

        # call parent method
        super().__init__(ops=fort_ops, imports=imports, file_ending='.f90', idx_left='(', idx_right=')', start_idx=1,
                         **kwargs)

        self._op_calls = {}
        self._var_declaration_info = {}

        # define fortran-specific imports
        self._imports.pop(0)
        self._imports.append("double precision :: PI = 4.0*atan(1.0)")

    def add_var_update(self, lhs: ComputeVar, rhs: str, lhs_idx: Optional[str] = None, rhs_shape: Optional[tuple] = ()):
        self.register_vars([lhs])
        super().add_var_update(lhs, rhs, lhs_idx, rhs_shape)

    def create_index_str(self, idx: Union[str, int, tuple], separator: str = ',', apply: bool = True,
                         **kwargs) -> Tuple[str, dict]:

        if not apply:
            self._start_idx = 0
            idx, idx_dict = super().create_index_str(idx, separator, apply, **kwargs)
            self._start_idx = 1
            return idx, idx_dict
        else:
            return super().create_index_str(idx, separator, apply, **kwargs)

    def generate_func_head(self, func_name: str, state_var: str = 'y', return_var: str = 'dy', func_args: list = None,
                           add_hist_func: bool = False):

        # finalize list of arguments for the function call
        if func_args:
            self.register_vars(func_args)
            func_args = [arg.name for arg in func_args]
        else:
            func_args = []
        state_vars = ['t', state_var]
        if add_hist_func:
            state_vars.append('hist')
        _, indices = np.unique(func_args, return_index=True)
        func_args = state_vars + [func_args[idx] for idx in np.sort(indices)]

        # define module
        self.add_code_line(f"module {self._fname}")
        self.add_linebreak()

        # add global variable definitions and other imports
        for imp in self._imports:
            self.add_code_line(imp)

        # start function definition section
        self.add_linebreak()
        self.add_code_line("contains")
        self.add_linebreak()

        # add function header
        self.add_linebreak()
        self._add_func_call(name=func_name, args=func_args, return_var=return_var)

        return func_args

    def generate_func_tail(self, rhs_var: str = None):

        # end the subroutine
        self.add_code_line(f"end subroutine")
        self.add_linebreak()
        self.add_linebreak()

        # add definitions of helper functions after the main run function
        for func in self._helper_funcs:
            self.add_code_line(func)
            self.add_linebreak()

        # end the module
        self.add_code_line(f"end module")

    def add_code_line(self, code_str):
        """Add code line string to code.
        """
        for code in code_str.split('\n'):
            if self.linebreak_end not in code:
                code = code.replace('\t', '')
                code = '\t' * self.lvl + code
            if self.break_line(code):
                idx = self._find_first_op(code, start=len(self.linebreak_start),
                                          stop=self.n2 - len(self.linebreak_end))
                self.add_code_line(f'{code[0:idx]}{self.linebreak_end}')
                code = f"{self.linebreak_start}{code[idx:]}"
                self.add_code_line(code)
            else:
                self.code.append(code)

    def break_line(self, code: str):
        n = len(code)
        if n > self.n2:
            return True
        if n > self.n1:
            if self.linebreak_start in code:
                if self.linebreak_end in code[len(self.linebreak_start):]:
                    if n - len(self.linebreak_start) - len(self.linebreak_end) < self.n2:
                        return False
                    return True
                if n - len(self.linebreak_start) < self.n2:
                    return False
                return True
            if self.linebreak_end in code:
                if n - len(self.linebreak_end) < self.n2:
                    return False
                return True
            return False
        return False

    def generate_func(self, func_name: str, to_file: bool = True, func_args: tuple = (), state_vars: tuple = (),
                      **kwargs):

        file = f'{self.fdir}/{self._fname}{self._fend}' if self.fdir else f'{self._fname}{self._fend}'

        # generate the final string representing the function file
        auto_compatible = kwargs.pop('auto', False)
        if auto_compatible:

            # case I: generate both the auto function and auto constants strings and write the constants string to file
            func_file, const_file = self._generate_auto_files(func_name=func_name, func_args=func_args,
                                                              state_vars=state_vars, **kwargs)

            # write auto constants to file
            build_dir = f"{self.fdir}/" if self.fdir else ""
            cname = kwargs.pop('auto_constants_file', 'ivp')
            try:
                with open(f'{build_dir}c.{cname}', 'wt') as cfile:
                    cfile.write(const_file)
            except FileNotFoundError:
                with open(f'{build_dir}c.{cname}', 'xt') as cfile:
                    cfile.write(const_file)

        else:

            # case II: generate a standard fortran function string
            func_file = self.generate()

        # write function to file
        with open(file, 'w') as f:
            f.writelines(func_file)
            f.close()

        # compile fortran function and write it to file
        subprocess.run(f"python -m numpy.f2py -c -m {self._fname} {file}", shell=True,
                       stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

        # import function from temporary file
        exec(f"from {self._fname} import {self._fname}", globals())
        exec(f"rhs_eval = {self._fname}.{func_name}", globals())
        rhs_eval = globals().pop('rhs_eval')

        # apply function decorator
        decorator = kwargs.pop('decorator', None)
        if decorator:
            decorator_kwargs = kwargs.pop('decorator_kwargs', dict())
            rhs_eval = decorator(rhs_eval, **decorator_kwargs)

        if not to_file:
            os.remove(file)

        return rhs_eval

    def register_vars(self, variables: list):
        for v in variables:
            if v.name not in self._var_declaration_info:
                self._var_declaration_info[v.name] = v

    def clear(self) -> None:
        """Removes all layers, variables and operations from graph. Deletes build directory.
        """

        # delete fortran-specific temporary files
        wdir = self.fdir if self.fdir else os.getcwd()
        for f in [f for f in os.listdir(wdir)]:
            if "cpython" in f and self._fname in f and f[-3:] == ".so":
                os.remove(f"{wdir}/{f}")
            elif f == 'c.ivp' or (f[:5] == 'fort.' and len(f) == 6):
                os.remove(f"{wdir}/{f}")
            elif f == f"{self._fname}.exe" or f == f"{self._fname}.mod" or f == f"{self._fname}.o":
                os.remove(f"{wdir}/{f}")

        # call parent method
        super().clear()

    @staticmethod
    def expr_to_str(expr: str, args: tuple):

        func = 'cshift('
        if func in expr:

            old_shift = f"{args[-1]}"
            new_shift = f"-{old_shift}"
            start = expr.find(func) + len(func)
            stop = expr[start:].find(')')
            old_expr = expr[start:start+stop]
            new_expr = replace(expr[start:start+stop], old_shift, new_shift)
            expr = replace(expr, old_expr, new_expr)

        return expr

    def _generate_auto_files(self, func_name: str, func_args: tuple = (), state_vars: tuple = (),
                             blocked_indices: tuple = (10, 15), **kwargs):

        # wrapper to the right-hand side evaluation function
        ####################################################

        # add function header that Auto-07p expects
        self.add_linebreak()
        self.add_linebreak()
        self.add_code_line("subroutine func(ndim,y,icp,args,ijac,dy,dfdu,dfdp)")
        self.add_linebreak()

        # load the module in which the pyrates function has been defined
        self.add_code_line(f"use {self._fname}")

        # declare auto-related variables
        dtype = self._get_dtype(self._var_declaration_info['y'].dtype)
        self.add_code_line("implicit none")
        self.add_code_line("integer, intent(in) :: ndim, icp(*), ijac")
        self.add_code_line(f"{dtype}, intent(in) :: y(ndim), args(*)")
        self.add_code_line(f"{dtype}, intent(out) :: dy(ndim)")
        self.add_code_line(f"{dtype}, intent(inout) :: dfdu(ndim,ndim), dfdp(ndim,*)")

        # extract parameters from args list
        increment = 1
        params, param_indices = [], []
        for i, arg in enumerate(func_args):
            idx = i + increment
            if blocked_indices[0] <= idx <= blocked_indices[1]:
                idx -= increment
                increment += blocked_indices[1] - blocked_indices[0]
                idx += increment
            params.append(f'args({idx})')
            param_indices.append(idx)

        # call the pyrates subroutine
        additional_args = f", {', '.join(params)}" if params else ""
        self.add_linebreak()
        self.add_code_line(f"call {func_name}(args(14), y, dy{additional_args})")
        self.add_linebreak()
        self.add_code_line("end subroutine func")
        self.add_linebreak()

        # routine that sets up an initial value problem
        ###############################################

        # generate subroutine header
        self.add_linebreak()
        self.add_code_line("subroutine stpnt(ndim, y, args, t)")
        self.add_linebreak()
        self.add_code_line("implicit None")
        self.add_code_line("integer, intent(in) :: ndim")
        self.add_code_line(f"{dtype}, intent(inout) :: y(ndim), args(*)")
        self.add_code_line(f"{dtype}, intent(in) :: t")
        self.add_linebreak()

        # define parameter values
        increment = 1
        for idx, arg in zip(param_indices, func_args):
            p = self._var_declaration_info[arg]
            if sum(p.shape) > 1:
                raise ValueError(f'Vector-valued parameter detected ({p.name} with shape {p.shape}), which cannot be '
                                 f'handled by Auto-07p. Please change the definition of your network (e.g. remove '
                                 f'extrinsic inputs) such that no vectorized model parameters exist.')
            self.add_code_line(f"args({idx}) = {p.value}  ! {p.name}")

        # define initial state
        for i, var in enumerate(state_vars):
            v = self._var_declaration_info[var]
            self.add_code_line(f"y({i+1}) = {v.value}  ! {v.name}")

        # end subroutine
        self.add_linebreak()
        self.add_code_line("end subroutine stpnt")
        self.add_linebreak()

        # dummy routines (could be made available for more complex Auto-07p usages)
        ###########################################################################

        self.add_linebreak()
        for routine in ['bcnd', 'icnd', 'fopt', 'pvls']:
            self.add_linebreak()
            self.add_code_line(f"subroutine {routine}")
            self.add_code_line(f"end subroutine {routine}")
            self.add_linebreak()

        func_file = self.generate()
        self.code.clear()

        # create auto constants file
        ############################

        auto_constants = {'NDIM': 1, 'NPAR': 1, 'IPS': -2, 'ILP': 0, 'ICP': [14], 'NTST': 1, 'NCOL': 3, 'IAD': 0,
                          'ISP': 0, 'ISW': 1, 'IPLT': 0, 'NBC': 0, 'NINT': 0, 'NMX': 10000, 'NPR': 1, 'MXBF': 10,
                          'IID': 2, 'ITMX': 2, 'ITNW': 5, 'NWTN': 2, 'JAC': 0, 'EPSL': 1e-6, 'EPSU': 1e-6, 'EPSS': 1e-4,
                          'IRS': 0, 'DS': 1e-4, 'DSMIN': 1e-8, 'DSMAX': 1e-2, 'IADS': 1, 'THL': {}, 'THU': {},
                          'UZR': {}, 'STOP': {}}

        auto_constants['NDIM'] = len(state_vars)
        auto_constants['NPAR'] = max(param_indices)
        for key in list(kwargs.keys()):
            if key in auto_constants:
                auto_constants[key] = kwargs.pop(key)

        # write auto constants to string
        for key, val in auto_constants.items():
            self.add_code_line(f"{key} = {val}")
        const_file = self.generate()
        self.code.clear()

        return func_file, const_file

    def _get_func_info(self, name: str, shape: tuple = (), dtype: str = 'float'):

        func_info = self._funcs[name]

        # case I: generate shape-specific fortran function call
        if callable(func_info['call']):

            # extract unique index for input variable shape
            try:
                shapes, indices = self._op_calls[name]
                try:
                    idx = shapes.index(shape)
                    idx = indices[idx]
                except IndexError:
                    idx = indices[-1]
                    shapes.append(shape)
                    indices.append(idx)
            except KeyError:
                idx = 1
                self._op_calls[name] = [shape], [idx]

            # generate function call and string
            func_call, func_str = func_info['call'](idx, self._get_shape(shape, var=''), self._get_dtype(dtype))
            func_info['call'] = func_call
            func_info['def'] = func_str

        return func_info

    def _add_func_call(self, name: str, args: Iterable, return_var: str = 'dy'):

        # add function header
        self.add_code_line(f"subroutine {name}({','.join(args)})")
        self.add_linebreak()
        self.add_code_line("implicit none")
        self.add_linebreak()

        # add variable declarations
        for arg in self._var_declaration_info:
            dtype, intent, shape = self._get_var_declaration_info(arg, args)
            intent = f", intent({intent})" if intent else ""
            self.add_code_line(f"{dtype}{intent} :: {arg}{shape}")

    def _get_var_declaration_info(self, var: str, args: Iterable) -> tuple:

        # extract variable definition
        v = self._var_declaration_info[var]

        # define data type
        dtype = self._get_dtype(v.dtype)

        # define intent of input arguments
        if v.name in args:
            intent = 'in' if v.is_constant or v.name in 'ty' else 'inout'
        else:
            intent = ""

        # define shape
        shape = self._get_shape(v.shape, var)

        return dtype, intent, shape

    def _solve(self, solver: str, func: Callable, args: tuple, T: float, dt: float, dts: float, y0: np.ndarray,
               t0: np.ndarray, times: np.ndarray, **kwargs) -> np.ndarray:

        # extract delta vector
        dy = args[0]

        # define wrapper function for fortran subroutine
        def fort_func(t, y, *args):
            func(t, y, *args)
            return dy

        return super()._solve(solver=solver, func=fort_func, args=args, T=T, dt=dt, dts=dts, y0=y0, t0=t0, times=times,
                              **kwargs)

    def _get_dtype(self, dtype: Union[str, np.dtype]):
        if dtype == 'float':
            dtype = self._float_precision
        if 'float' in dtype:
            dtype = 'double precision' if '64' in dtype else 'real'
        elif 'complex' in dtype:
            dtype = 'complex'
        else:
            dtype = 'integer'
        return dtype

    def _process_idx(self, idx: Union[Tuple[int, int], int, str, ComputeVar], **kwargs) -> str:
        if idx == ':':
            return ''
        return super()._process_idx(idx=idx, **kwargs)

    @staticmethod
    def _get_shape(shape: tuple, var: str):
        shape = str(shape) if shape else ''
        if len(shape) < 3:
            shape = '(1)' if (var == 'dy' or var == 'y') else ''
        elif shape[-2] == ',':
            shape = f"{shape[:-2]})"
        return shape

    @staticmethod
    def _find_first_op(code, start, stop):
        if stop < len(code):
            code_tmp = code[start:stop]
            ops = ["+", "-", "*", "/", "**", "^", "%", "<", ">", "==", "!=", "<=", ">="]
            indices = [code_tmp.index(op) for op in ops if op in code_tmp]
            if indices and max(indices) > 0:
                return max(indices) + start
            idx = start
            for break_sign in [',', ')', ' ']:
                if break_sign in code_tmp:
                    idx_tmp = len(code_tmp) - code_tmp[::-1].index(break_sign)
                    if len(code_tmp) - idx_tmp < len(code_tmp) - idx:
                        idx = idx_tmp
            return idx + start
        return stop + start



"""Interface to load from and save to python pickles."""


def dump(data, filename, **kwargs):
    import pickle

    from pyrates.utility import create_directory
    create_directory(filename)

    pickle.dump(data, open(filename, 'wb'), protocol=pickle.HIGHEST_PROTOCOL, **kwargs)


def load(filename, **kwargs):
    import pickle

    data = pickle.load(open(filename, 'rb'), **kwargs)

    return data



def benchmark_single_jr_circuit(n_runs=10):
    """This is a simple benchmark function that instantiates a basic JR circuit and compiles it.

    Parameters
    ----------
    n_runs
        number of runs to execute benchmark for averaging
    """

    import time
    from pyrates import CircuitTemplate
    from pyrates import clear
    import numpy as np

    path = "model_templates.neural_mass_models.jansenrit.JRC"

    timer_dict = dict(template_load=[],
                      template_apply=[],
                      # circuit_compile=[]
                      )

    for i in range(n_runs):

        # time template loading
        tic = time.perf_counter()
        template = CircuitTemplate.from_yaml(path)
        toc = time.perf_counter()
        timer_dict["template_load"].append(toc - tic)

        # time template application to circuit IR
        tic = time.perf_counter()
        template.apply(verbose=False)
        toc = time.perf_counter()
        timer_dict["template_apply"].append(toc - tic)

        clear(template)

    print("Benchmark results:")
    for key, times in timer_dict.items():
        median = np.median(times)

        print(f"{key}: {median:.4f}s")


if __name__ == "__main__":
    benchmark_single_jr_circuit(10)




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network models and simulations. See also: 
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
""" Tests for the parser that translates circuits and components defined in YAML into the intermediate
python representation.
"""

__author__ = "Daniel Rose"
__status__ = "Development"

import pytest
import numpy as np


def setup_module():
    print("\n")
    print("===========================")
    print("| Test Suite: YAML Parser |")
    print("===========================")


@pytest.mark.parametrize("operator", ["model_templates.base_templates.li_op",
                                      "model_templates.base_templates.alpha_op",
                                      "model_templates.base_templates.sigmoid_op",
                                      "model_templates.neural_mass_models.jansenrit.rpo_e_in",
                                      "model_templates.neural_mass_models.qif.qif_sfa_op",
                                      "model_templates.oscillators.kuramoto.sin_op"
                                      ])
def test_import_operator_templates(operator):
    """test basic (vanilla) YAML parsing using ruamel.yaml (for YAML 1.2 support)"""
    from pyrates import OperatorTemplate
    from pyrates.frontend.template import template_cache, clear_cache
    clear_cache()

    template = OperatorTemplate.from_yaml(operator)  # type: OperatorTemplate

    assert template.path in template_cache

    cached_template = template_cache[operator]  # type: OperatorTemplate
    assert template is cached_template
    assert template.path == cached_template.path
    assert template.equations == cached_template.equations
    assert template.variables == cached_template.variables
    assert repr(template) == repr(cached_template) == f"<OperatorTemplate '{operator.split('.')[-1]}'>"


def test_full_jansen_rit_circuit_template_load():
    """Test a simple circuit template, including all nodes, edges and operators to be loaded."""

    path = "model_templates.neural_mass_models.jansenrit.JRC"
    from pyrates.frontend.template.circuit import CircuitTemplate
    from pyrates.frontend.template.node import NodeTemplate
    from pyrates.frontend.template.operator import OperatorTemplate
    from pyrates.frontend.template import template_cache, clear_cache
    clear_cache()

    template = CircuitTemplate.from_yaml(path)

    # test, whether circuit is in loader cache
    assert template is template_cache[path]

    # test, whether node templates have been loaded successfully
    nodes = {"pc": "model_templates.neural_mass_models.jansenrit.PC",
             "ein": "model_templates.neural_mass_models.jansenrit.IN",
             "iin": "model_templates.neural_mass_models.jansenrit.IN"}

    for key, value in nodes.items():
        assert isinstance(template.nodes[key], NodeTemplate)
        assert template.nodes[key] is template_cache[value]
        # test operators in node templates
        for op in template.nodes[key].operators:
            assert op.path in template_cache
            assert isinstance(op, OperatorTemplate)

    # test that all item views work correctly
    for key, value in nodes.items():
        assert template[key] is template_cache[value]
        assert type(template[key]["pro"]) is OperatorTemplate
        assert "input" in template[key]["pro"]["v"]


def test_edge_definition_via_matrix():
    """Test, if CircuitTemplate.add_edges_from_matrix works as expected."""

    from pyrates import CircuitTemplate, NodeTemplate, EdgeTemplate
    import numpy as np
    from copy import deepcopy

    # define the network without edges
    node = NodeTemplate.from_yaml("model_templates.oscillators.kuramoto.phase_pop")
    node_names = ['p1', 'p2', 'p3']
    n = len(node_names)
    circuit = CircuitTemplate(name='delay_coupled_kmos', nodes={key: node for key in node_names})

    # add the edges
    edge = EdgeTemplate.from_yaml("model_templates.oscillators.kuramoto.sin_edge")
    weights = np.random.randn(n, n)
    delays = np.random.uniform(low=1, high=2, size=(n, n))
    edge_attr = {'sin_edge/coupling_op/theta_s': 'source', 'sin_edge/coupling_op/theta_t': 'p2/phase_op/theta',
                 'delay': delays}
    circuit.add_edges_from_matrix(source_var='phase_op/theta', target_var='phase_op/s_in', source_nodes=node_names,
                                  weight=weights, template=edge, edge_attr=edge_attr, min_weight=0.0)

    # test whether edges have been added as expected
    edge_attr_tmp = deepcopy(edge_attr)
    edge_attr_tmp['weight'] = weights[1, 2]
    edge_attr_tmp['delay'] = delays[1, 2]
    assert len(circuit.edges) == int(n**2)
    assert ('p3/phase_op/theta', 'p2/phase_op/s_in', edge, edge_attr_tmp) in circuit.edges

    # perform short simulation to ensure that the network has been constructed correctly
    circuit.run(simulation_time=1.0, step_size=1e-4, outputs=['all/phase_op/theta'])


def test_circuit_instantiation():
    """Test, if apply() functions all work properly"""
    path = "model_templates.neural_mass_models.jansenrit.JRC"
    from pyrates import clear_frontend_caches
    from pyrates.frontend import template
    from pyrates.backend.computegraph import ComputeGraph, ComputeVar
    clear_frontend_caches()

    circuit = template.from_yaml(path)

    circuit.apply()
    ir = circuit.intermediate_representation

    # test whether calling apply translated the template into a proper intermediate representation
    assert type(ir.graph) is ComputeGraph
    assert circuit._vectorization_indices['pc/pro/s'] == [0]
    assert len(np.unique(list(circuit._vectorization_indices.values()))) == 2
    assert len(circuit._vectorization_labels) == 3
    assert type(ir._front_to_back['pc/pro/m']) is ComputeVar


def test_multi_circuit_instantiation():
    """Test, if a circuit with subcircuits is also working."""
    path = "model_templates.neural_mass_models.jansenrit.JRC_2delaycoupled"
    from pyrates import clear_frontend_caches, CircuitTemplate
    clear_frontend_caches()

    circuit = CircuitTemplate.from_yaml(path)
    assert circuit


def test_equation_alteration():
    """Test, if properties of a template that mean to alter a certain parent equation are treated correctly"""

    path = "model_templates.neural_mass_models.jansenrit.rpo_e_in"
    # template replaces the component "m_in" with "(m_in + u)" in the equation "X' = h*m_in/tau - 2*X/tau - V/tau**2"
    from pyrates.frontend.template.operator import OperatorTemplate

    template = OperatorTemplate.from_yaml(path)
    operator, _ = template.apply()

    assert operator.equations[1] == "x' = h*(m_in + u)/tau - 2*x/tau - v/tau**2"


def test_python_interface():
    """Test, if yaml and Python interface for template definition deliver equivalent results.
    """
    from pyrates import OperatorTemplate, NodeTemplate, CircuitTemplate, clear
    import numpy as np

    # build qif model via python interface
    eqs = [
        "r' = (Delta/(pi*tau) + 2.0*r*v) / tau",
        "v' = (v^2 + eta + I_ext + tau*r_in - (pi*tau*r)^2) / tau"
    ]

    variables = {
        "r": "output(0.01)",
        "v": "variable(-2.0)",
        "Delta": 1.0,
        "tau": 1.0,
        "eta": -5.0,
        "I_ext": "input(0.0)",
        "r_in": "input(0.0)"
    }

    op = OperatorTemplate(name='qif_op', equations=eqs, variables=variables, path=None)
    node = NodeTemplate(name='qif_pop', operators=[op], path=None)
    qif_python = CircuitTemplate(name='qif', nodes={'p': node},
                                 edges=[('p/qif_op/r', 'p/qif_op/r_in', None, {'weight': 15.0})]
                                 )

    # load YAML-based qif model definition from template
    qif_yaml = CircuitTemplate.from_yaml('model_templates.neural_mass_models.qif.qif')

    # compare dynamics of QIF population defined above with QIF population defined via YAML interface
    T = 40.0
    dt = 1e-3
    dts = 1e-2

    in_start = int(np.round(10.0 / dt))
    in_dur = int(np.round(20.0 / dt))
    inp = np.zeros((int(np.round(T / dt)),))
    inp[in_start:in_start + in_dur] = 3.0

    r1 = qif_python.run(simulation_time=T, step_size=dt, sampling_step_size=dts, solver='scipy',
                        outputs={'r': 'p/qif_op/r'}, inputs={'p/qif_op/I_ext': inp})
    clear(qif_python)
    r2 = qif_yaml.run(simulation_time=T, step_size=dt, sampling_step_size=dts, solver='scipy',
                      outputs={'r': 'p/qif_op/r'}, inputs={'p/qif_op/I_ext': inp})
    clear(qif_yaml)

    assert np.mean(r1.values - r2.values) == pytest.approx(0.0, rel=1e-4, abs=1e-4)



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

"""This module provides parser classes and functions to parse string-based equations into symbolic representations of
operations.
"""

# external _imports
import math
import typing as tp
from numbers import Number

import numpy as np
# from pyparsing import Literal, CaselessLiteral, Word, Combine, Optional, \
#     ZeroOrMore, Forward, nums, alphas, ParserElement
from sympy import Expr, Symbol, lambdify, sympify

# pyrates internal _imports
from pyrates.backend.computegraph import ComputeGraph, ComputeNode

# meta infos
__author__ = "Richard Gast"
__status__ = "development"

###############################################
# expression parsers (lhs/rhs of an equation) #
###############################################


class Algebra:

    def __init__(self, **kwargs) -> None:
        """Instantiates expression parser.
        """

        # call super method
        super().__init__()
        self.algebra = kwargs.pop('algebra', None)

        # define algebra
        ################

        # if not self.algebra:
        #
        #     # general symbols
        #     point = Literal(".")
        #     comma = Literal(",")
        #     colon = Literal(":")
        #     e = CaselessLiteral("E")
        #     pi = CaselessLiteral("PI")
        #
        #     # parentheses
        #     par_l = Literal("(")
        #     par_r = Literal(")").setParseAction(self._push_first)
        #     idx_l = Literal("[")
        #     idx_r = Literal("]")
        #
        #     # basic mathematical operations
        #     plus = Literal("+")
        #     minus = Literal("-")
        #     mult = Literal("*")
        #     div = Literal("/")
        #     mod = Literal("%")
        #     dot = Literal("@")
        #     exp_1 = Literal("^")
        #     exp_2 = Combine(mult + mult)
        #     transp = Combine(point + Literal("T"))
        #     inv = Combine(point + Literal("I"))
        #
        #     # numeric types
        #     num_float = Combine(Word("-" + nums, nums) +
        #                         Optional(point + Optional(Word(nums))) +
        #                         Optional(e + Word("-" + nums, nums)))
        #     num_int = Word("-" + nums, nums)
        #
        #     # variables and functions
        #     name = Word(alphas, alphas + nums + "_$")
        #     func_name = Combine(name + par_l, adjacent=True)
        #
        #     # math operation groups
        #     op_add = plus | minus
        #     op_mult = mult | div | dot | mod
        #     op_exp = exp_1 | exp_2 | inv | transp
        #
        #     # logical operations
        #     greater = Literal(">")
        #     less = Literal("<")
        #     equal = Combine(Literal("=") + Literal("="))
        #     unequal = Combine(Literal("!") + Literal("="))
        #     greater_equal = Combine(Literal(">") + Literal("="))
        #     less_equal = Combine(Literal("<") + Literal("="))
        #
        #     # logical operations group
        #     op_logical = greater_equal | less_equal | unequal | equal | less | greater
        #
        #     # pre-allocations
        #     self.algebra = Forward()
        #     exponential = Forward()
        #     index_multiples = Forward()
        #
        #     # basic organization units
        #     index_start = idx_l.setParseAction(self._push_first)
        #     index_end = idx_r.setParseAction(self._push_first)
        #     index_comb = colon.setParseAction(self._push_first)
        #     arg_comb = comma.setParseAction(self._push_first)
        #     arg_tuple = par_l + ZeroOrMore(self.algebra.suppress() + Optional(arg_comb)) + par_r
        #     func_arg = arg_tuple | self.algebra.suppress()
        #
        #     # basic computation unit
        #     atom = (func_name + Optional(func_arg.suppress()) + ZeroOrMore(arg_comb.suppress() + func_arg.suppress()) +
        #             par_r.suppress() | name | pi | e | num_float | num_int).setParseAction(self._push_neg_or_first) | \
        #            (par_l.setParseAction(self._push_last) + self.algebra.suppress() + par_r).setParseAction(self._push_neg)
        #
        #     # apply indexing to atoms
        #     indexed = (Optional(minus) + atom).setParseAction(self._push_neg) + \
        #               ZeroOrMore((index_start + index_multiples + index_end))
        #     index_base = (self.algebra.suppress() | index_comb)
        #     index_full = index_base + ZeroOrMore((index_comb + index_base)) + ZeroOrMore(index_comb)
        #     index_multiples << index_full + ZeroOrMore((arg_comb + index_full))
        #
        #     # hierarchical relationships between mathematical and logical operations
        #     boolean = indexed + Optional((op_logical + indexed).setParseAction(self._push_first))
        #     exponential << boolean + ZeroOrMore((op_exp + Optional(exponential)).setParseAction(self._push_first))
        #     factor = exponential + ZeroOrMore((op_mult + exponential).setParseAction(self._push_first))
        #     expr = factor + ZeroOrMore((op_add + factor).setParseAction(self._push_first))
        #     self.algebra << expr

    def _push_first(self, strg, loc, toks):
        """Push tokens in first-to-last order to expression stack.
        """
        self.expr_stack.append(toks[0])

    def _push_neg(self, strg, loc, toks):
        """Push negative one multiplier if on first position in toks.
        """
        if toks and toks[0] == '-':
            self.expr_stack.append('-one')

    def _push_neg_or_first(self, strg, loc, toks):
        """Push neg one multipler to expression stack if on first position in toks, else push toks from first-to-last.
        """
        if toks and toks[0] == '-':
            self.expr_stack.append('-one')
        else:
            self.expr_stack.append(toks[0])

    def _push_last(self, strg, loc, toks):
        """Push tokens in last-to-first order to expression stack.
        """
        self.expr_stack.append(toks[-1])


class ExpressionParser:
    """Class for parsing mathematical expressions from a string format into a symbolic representation of the
    mathematical operation expressed by it.

    Parameters
    ----------
    expr_str
        Mathematical expression in string format.
    args
        Dictionary containing all variables and functions needed to evaluate the expression.
    cg
        ComputeGraph instance in which to parse all variables and operations.
        See `pyrates.backend.computegraph.ComputeGraph` for a full documentation of its methods and attributes.
    parsing_method
        Name of the parsing method to use. Valid options: `sympy` for a sympy-based parser, `pyrates` for the
        pyrates-internal, pyparsing-based parser.

    Attributes
    ----------
    lhs
        Boolean, indicates whether expression is left-hand side or right-hand side of an equation
    rhs
        PyRatesOp for the evaluation of the right-hand side of the equation
    vars
        Dictionary containing the variables of an expression
    expr_str
        String representation of the mathematical expression
    expr_stack
        List representation of the syntax tree of the (parsed) mathematical expression.
    parse_func
        Function that will be used to deduce an operation stack/tree from a given expression.
    cg
        Passed ComputeGraph instance into which the expression will be parsed.

    """

    _constant_counter = 0

    def __init__(self, expr_str: str, args: dict, cg: ComputeGraph, def_shape: tp.Optional[tuple] = None,
                 parsing_method: str = 'sympy') -> None:
        """Instantiates expression parser.
        """

        # main attributes
        self.vars = args.copy()
        self.expr_str = expr_str
        self.expr_stack = []
        self.cg = cg
        self._def_shape = def_shape

        # preprocess the expression string
        self.lhs, self.rhs, self._diff_eq, self._assign_type, self.lhs_key = self._preprocess_expr_str(expr_str)

        # define the parsing function
        if parsing_method == 'sympy':
            self.parse_func = sympify
        elif parsing_method == 'pyrates':
            a = Algebra()
            self.parse_func = a.parseString
        else:
            raise ValueError(f'Invalid identifier for the parsing method: {parsing_method}.')

    def parse_expr(self) -> dict:
        """Parses string-based mathematical expression/equation.

        Returns
        -------
        dict
            Variables of the parsed equation.
        """

        # extract symbols and operations from equations right-hand side
        self.expr_stack = self.parse_func(self.rhs)
        if self.expr_stack.is_number:
            c = f"dummy_constant_{self._constant_counter}"
            expr = f"no_op({c})"
            self.vars[c] = {'vtype': 'input',
                            'value': float(self.expr_stack),
                            'shape': ()}
            self.expr_stack = self.parse_func(expr)
            self._constant_counter += 1
        if self.expr_stack.is_symbol:
            self.expr_stack = self.parse_func(f"no_op({self.rhs})")

        # parse rhs into backend
        self.rhs = self._parse_stack(self.expr_stack)

        # extract symbols and operations from left-hand side
        self.expr_stack = self.parse_func(self.lhs)

        # parse lhs into backend
        self._update_lhs()

        return self.vars
    
    def _parse_stack(self, expr: Expr):

        if expr.args:

            # parse variables as nodes into compute graph
            inputs, func_args, old_args = [], [], []
            for arg in self._sort_expr_args(expr.args):

                if isinstance(arg, Symbol):

                    # case I: variables/constants
                    var = self.vars[arg.name]
                    if isinstance(var, ComputeNode):

                        # if parsed already, retrieve label from existing variable
                        label = var.name

                    else:

                        # if not parsed already, parse variable into backend
                        label, var = self.cg.add_var(label=arg.name, def_shape=self._def_shape, **var)
                        self.vars[arg.name] = var

                else:

                    # case II: mathematical expressions
                    label, var = self._parse_stack(arg)

                if isinstance(var, ComputeNode):

                    # store input to mathematical expression, if it is not a simple scalar
                    inputs.append(label)
                    func_args.append(var.symbol)
                    old_args.append(arg)

            # replace names of old expression arguments with new variable symbols
            replacements = {old: new for old, new in zip(old_args, func_args) if old != new}
            if replacements:
                expr = replace_in_expr(expr, replacements)

            # create callable function of the operation
            label = expr.func.__name__
            try:
                v_tmp = self.cg.get_var(func_args[0].name)
                op = self.cg.get_op(label, shape=v_tmp.shape, dtype=v_tmp.dtype)
                backend_funcs = {label: op['func']}
            except (KeyError, IndexError):
                backend_funcs = dict()
            func = lambdify(func_args, expr=expr, modules=[backend_funcs, "numpy"])

            # parse mathematical operation into compute graph
            return self.cg.add_op(inputs, label=label, expr=expr, func=func)

        else:

            # for simple scalar constants, return empty string and dict
            return "", dict()
        
    def _preprocess_expr_str(self, expr: str) -> tuple:
        """Turns differential equations into simple algebraic equations using a certain solver scheme and extracts
        left-hand side, right-hand side and update type of the equation.

        Parameters
        ----------
        expr
            Equation in string format.

        Returns
        -------
        tuple
            Contains left hand side, right hand side and left hand side update type
        """

        # collect equation specifics
        ############################

        # split equation into lhs and rhs and assign type
        lhs, rhs, assign_type = split_equation(expr)

        if not assign_type:
            self.vars['x'] = {'vtype': 'variable', 'value': 0.0, 'dtype': 'float', 'shape': ()}
            return self._preprocess_expr_str(f"x = {expr}")

        # for the left-hand side, check whether it includes a differential operator
        if "d/dt" in lhs:
            diff_eq = True
            lhs_split = lhs.split('*')
            lhs = "".join(lhs_split[1:])
        elif "'" in lhs:
            diff_eq = True
            lhs = lhs.replace("'", "")
        elif "d" in lhs and "/dt" in lhs:
            diff_eq = True
            lhs = lhs.split('/dt')[0]
            lhs = lhs.replace("d", "", count=1)
        else:
            diff_eq = False

        # get clean name of lhs
        lhs_key = lhs.split('(')[0]
        lhs_key = lhs_key.replace(' ', '')
        lhs = lhs.replace(' ', '')

        # store equation specifics
        if diff_eq and assign_type != '=':
            raise ValueError(f'Wrong assignment method for equation: {expr}. '
                             f'A differential equation cannot be combined with an assign type other than `=`.')

        return lhs, rhs, diff_eq, assign_type, lhs_key
    
    def _update_lhs(self):
        """Applies update to left-hand side of equation. For differential equations, different solving schemes are
        available.
        """

        # update left-hand side of equation
        ###################################

        # receive left-hand side variable information
        lhs_key = self.lhs_key
        if self.expr_stack.is_symbol:

            # retrieve variable information
            v = self.vars[self.lhs_key]

            # create backend state variable if it does not exist already
            if not isinstance(v, ComputeNode):
                _, v = self.cg.add_var(label=lhs_key, def_shape=self._def_shape, **v)

        else:

            # parse left-hand side indexing operation
            lhs_key, v = self._parse_stack(self.expr_stack)

        # create mapping between left-hand side and right-hand side of the equation
        self.cg.add_var_update(v.name, self.rhs[0], differential_equation=self._diff_eq)
        if lhs_key in self.vars:
            self.vars[lhs_key] = v

    @staticmethod
    def _sort_expr_args(args: tuple) -> list:

        # sort arguments from longest to shortest expression
        arg_lengths = [len(str(arg)) for arg in args]
        args_sorted = [args[idx] for idx in np.argsort(arg_lengths)[::-1]]

        # add arguments that need to be treated with priority
        args_final = []
        while args_sorted:

            arg = args_sorted.pop(0)

            # check whether the position between arguments should be swapped
            prioritize = False
            if isinstance(arg, Expr):
                for arg_tmp in args_sorted:
                    if len(arg_tmp.find(arg)):
                        prioritize = True
                        break

            if prioritize:
                idx = args_sorted.index(arg_tmp)
                args_final.append(args_sorted.pop(idx))
            else:
                args_final.append(arg)

        return args_final

################################
# helper classes and functions #
################################


def parse_equations(equations: list, equation_args: dict, cg: ComputeGraph, def_shape: tuple, **kwargs) -> dict:
    """Parses a system (list) of equations into the backend. Transforms differential equations into the appropriate set
    of right-hand side evaluations that can be solved later on.

    Parameters
    ----------
    equations
        Collection of equations that describe the dynamics of the nodes and edges.
    equation_args
        Key-value pairs of arguments needed for parsing the equations.
    cg
        ComputeGraph instance that all equations will be parsed into.
    def_shape
        Default shape of variables that are scalar. Can either be `(1,)` or `()`.
    kwargs
        Additional keyword arguments to be passed to the backend methods.

    Returns
    -------
    dict
        The updated equations args (in-place manipulation of all variables in equation_args happens during
        equation parsing).

    """

    for eq, scope in equations:

        scope = scope.split('/')

        # parse arguments
        #################

        # extract operator variables from equation args
        op_args = {}
        in_vars = []
        update_vars = {}
        for key, var in equation_args.copy().items():

            key_split = key.split('/')

            if key_split[:-1] == scope:

                var_name = key_split[-1]

                if var_name == 'inputs':

                    # extract inputs from other variable scopes
                    for in_key, inp in var.items():

                        # check whether input variable has been passed properly
                        if inp not in equation_args:
                            raise KeyError(inp)

                        # extract input variable
                        inp_tmp = equation_args[inp]

                        # add input variable to operator arguments
                        op_args[in_key] = inp_tmp

                        # remember to update the variable entry in the variable collection later
                        update_vars[inp] = in_key

                        # add variable to operator inputs
                        in_vars.append(in_key)

                elif var_name not in in_vars:

                    # change the variable type of input variables that did not receive any extrinsic input
                    try:
                        if var['vtype'] == 'input' and var_name not in equation_args['inputs']:
                            var['vtype'] = 'constant'
                    except KeyError:
                        var['vtype'] = 'constant'
                    except TypeError:
                        pass

                    # include variable information in operator arguments
                    op_args[var_name] = var

                    # remember to update the variable entry in the variable collection later
                    update_vars[key] = var_name

        # parse equation
        ################

        instantaneous = is_diff_eq(eq) is False

        # initialize parser
        parser = ExpressionParser(expr_str=eq, args=op_args, cg=cg, def_shape=def_shape, **kwargs)

        # parse expression into compute graph
        variables = parser.parse_expr()

        # store newly created backend variables
        for full_key, var_key in update_vars.items():
            if full_key in equation_args:
                equation_args[full_key] = variables[var_key]

    return equation_args


def split_equation(expr: str) -> tuple:
    """Splits an equation string into a left-hand side, right-and side and an assign type.

    Parameters
    ----------
    expr
        Equation string. Should contain a left-hand side and a right-hand side, separated by some form of assign symbol.

    Returns
    -------
    tuple
        left-hand side string, right-hand side string, assign operation string.
    """

    # define assign types and explicit non-assign types
    assign_types = ['+=', '-=', '*=', '/=']
    not_assign_types = ['<=', '>=', '==', '!=']

    lhs, rhs, assign_type, found_assign_type = "", "", "", False

    # look for assign types in expression
    for assign_type in assign_types:

        if assign_type in expr:

            # split expression via assign symbol
            if f' {assign_type} ' in expr:
                lhs, rhs = expr.split(f' {assign_type} ', maxsplit=1)
            elif f' {assign_type}' in expr:
                lhs, rhs = expr.split(f' {assign_type}', maxsplit=1)
            elif f'{assign_type} ' in expr:
                lhs, rhs = expr.split(f'{assign_type} ', maxsplit=1)
            else:
                lhs, rhs = expr.split(assign_type, maxsplit=1)
            found_assign_type = True
            break

        elif '=' in expr:

            # assume standard assign
            assign_type = '='
            assign = True

            # check if `=` symbol marks an assign operation or not
            for not_assign_type in not_assign_types:
                if not_assign_type in expr:
                    expr_tmp = expr.replace(not_assign_type, '')
                    if '=' not in expr_tmp:
                        assign = False

            if assign:

                # split equation via `=` symbol
                if f' = ' in expr:
                    lhs, rhs = expr.split(f' = ', maxsplit=1)
                elif f' {assign_type}' in expr:
                    lhs, rhs = expr.split(f' =', maxsplit=1)
                elif f'{assign_type} ' in expr:
                    lhs, rhs = expr.split(f'= ', maxsplit=1)
                else:
                    lhs, rhs = expr.split(f"=", maxsplit=1)
                found_assign_type = True
                break

    if not found_assign_type:
        return lhs, rhs, False
    return lhs, rhs, assign_type


def replace(eq: str, term: str, replacement: str, rhs_only: tp.Optional[bool] = False,
            lhs_only: tp.Optional[bool] = False) -> str:
    """Replaces a term in an equation with a replacement term (save replacement).

    Parameters
    ----------
    eq
        Equation that includes the term.
    term
        Term that should be replaced.
    replacement
        Replacement for all occurences of term.
    rhs_only
        If True, replacements will only be performed in right-hand side of the equation.
    lhs_only
        IF True, replacements will only be performed in left-hand side of the equation.

    Returns
    -------
    str
        The updated equation.
    """

    # define follow-up operations/signs that are allowed to follow directly after term in eq
    allowed_follow_ops = '-+=*/^<>=!.%@[]():, '

    # replace every proper appearance of term in eq with replacement
    ################################################################

    eq_new = ""
    idx = eq.find(term)

    # go through all appearances of term in eq
    while idx != -1:

        # get idx of sign that follows after term
        idx_follow_op = idx+len(term)

        # if it is an allowed sign, replace term, else not
        replaced = False
        if ((idx_follow_op < len(eq) and eq[idx_follow_op] in allowed_follow_ops) and
           (idx == 0 or eq[idx-1] in allowed_follow_ops)) or \
                (idx_follow_op == len(eq) and eq[idx-1] in allowed_follow_ops):
            eq_part = eq[:idx]
            if (rhs_only and "=" in eq_part) or (lhs_only and "=" not in eq_part) or (not rhs_only and not lhs_only):
                eq_new += f"{eq_part}{replacement}"
                replaced = True
        if not replaced:
            eq_new += f"{eq[:idx_follow_op]}"

        # jump to next appearance of term in eq
        eq = eq[idx_follow_op:]
        idx = eq.find(term)

    # add rest of eq to new eq
    eq_new += eq

    return eq_new


def is_diff_eq(eq: str) -> bool:
    """Checks whether `eq` is a differential equation or not.

    Parameters
    ----------
    eq
        Equation string.

    Returns
    -------
    bool
        True, if `eq` is a differential equation.

    """

    lhs, rhs, _ = split_equation(eq)
    if "d/dt" in lhs:
        de = True
    elif "'" in lhs:
        de = True
    elif "dt_test" in replace(rhs, "step_size", "dt_test"):
        de = True
    else:
        de = False

    return de


def var_in_expression(var: str, expr: str) -> bool:

    if var == expr:
        return True

    # define follow-up operations/signs that are allowed to follow directly after term in eq
    follow_ops = '+-=*/^<>=!.%@[]():, '

    start = 0
    n = len(expr)
    exists_in_expr = False
    while start < n:

        try:

            # find variable in string and extract the follow-up sign
            idx = expr[start:].find(var)
            if idx == -1:
                break
            idx += start
            idx_next = idx + len(var)
            next_sign = expr[idx_next]
            prev_sign = expr[idx - 1]

            # decide whether variable actually exists in expression
            exists_in_expr = ((idx_next < n and next_sign in follow_ops) and
                              (idx == 0 or prev_sign in follow_ops)) or (idx_next == n and prev_sign in follow_ops)

            if exists_in_expr:
                break
            start = idx_next

        except IndexError:

            break

    return exists_in_expr


def extract_var(var: str) -> tuple:
    if "[" in var:
        return var.split("[")[0], True
    return var, False


def get_unique_label(label: str, labels: dict) -> tp.Tuple[str, dict]:
    if label in labels:
        n = labels[label]
        if n == 0:
            label_new = f"{label}_num1"
        else:
            label_new = f"{label}_num{n+1}"
        labels[label] += 1
    else:
        label_new = label
        labels[label] = 0
    return label_new, labels


def replace_in_expr(expr: Expr, replacements: dict):
    expr = expr.subs(replacements, simultaneous=True)
    for arg_old in replacements:
        if expr.count(arg_old):
            expr = expr.replace(arg_old, replacements[arg_old])
    return expr




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network models and simulations. See also: 
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
""" Abstract base classes
"""

__author__ = "Daniel Rose"
__status__ = "Development"


class AbstractBaseTemplate:
    """Abstract base class for templates"""

    target_ir = None  # placeholder for template-specific intermediate representation (IR) target class

    def __init__(self, name: str, path: str = None, description: str = "A template."):
        """Basic initialiser for template classes, requires template name and path that it is loaded from. For custom
        templates that are not loaded from a file, the path can be set arbitrarily."""
        self.name = name
        self.path = path
        self.__doc__ = description  # overwrite class-specific doc with user-defined description

    def __repr__(self):
        """Defines how an instance identifies itself when called with `str()` or `repr()`, e.g. when shown in an
        interactive terminal. Shows Class name and path that was used to construct the class."""
        return f"<{self.__class__.__name__} '{self.name}'>"

    def __getitem__(self, item):
        raise NotImplementedError

    @classmethod
    def from_yaml(cls, path):
        """Load a template from yaml file. After importing the template, this method also checks whether
        the resulting template is actually an instance of the class that this method was called from. This is done to
        ensure any cls.from_yaml() produces only instances of that class and not other classes for consistency.
        Templates are cached by path. Depending on the 'base' key of the yaml template,
        either a template class is instantiated or the function recursively loads base templates until it hits a known
        template class.

        Parameters
        ----------
        path
            (str) path to YAML template of the form `path.to.template_file.template_name` or
            path/to/template_file/template_name.TemplateName. The dot notation refers to a path that can be found
            using python's import functionality. That means it needs to be a module (a folder containing an
            `__init__.py`) located in the Python path (e.g. the current working directory). The slash notation refers to
            a file in an absolute or relative path from the current working directory. In either case the second-to-last
            part refers to the filename without file extension and the last part refers to the template name.

        Returns
        -------

        """
        from pyrates.frontend.template import from_yaml
        tpl = from_yaml(path)

        if isinstance(tpl, cls):
            return tpl
        else:
            raise TypeError(f"The template associated with '{path}' is not of type {cls}.")

    def to_yaml(self, path, **kwargs) -> None:
        """Saves template to YAML file."""
        raise NotImplementedError

    def update_template(self, *args, **kwargs):
        """Updates the template with a given list of arguments and returns a new instance of the template class."""
        raise NotImplementedError

    def apply(self, *args, **kwargs):
        """Converts the template into its corresponding intermediate representation class."""
        raise NotImplementedError





# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""
"""
from typing import Iterator, Dict, List

import numpy as _np
from networkx import DiGraph, find_cycle, NetworkXNoCycle

from pyrates.backend import PyRatesException
from pyrates.ir.operator import OperatorIR

__author__ = "Daniel Rose"
__status__ = "Development"


class OperatorGraph(DiGraph):
    """Intermediate representation for nodes and edges."""

    def __init__(self, operators: Dict[str, OperatorIR] = None, template: str = ""):

        super().__init__()
        if operators is None:
            operators = {}

        # compute hash from incoming operators. Different order of operators in input might lead to different hash.
        self._h = hash(tuple(operators.values()))

        # collect all information about output variables of operators
        #############################################################
        all_outputs = {}  # type: Dict[str, List[str]]
        for key, operator in operators.items():

            # add operator as node to local operator_graph
            inputs = {var: dict(sources=set(), reduce_dim=True) for var in operator.inputs}
            self.add_node(key, operator=operator, inputs=inputs, label=key)

            # collect all output variables
            out_var = operator.output

            # check, if variable name exists in outputs and create empty list if it doesn't
            if out_var not in all_outputs:
                all_outputs[out_var] = []

            all_outputs[out_var].append(key)

        # link outputs to inputs
        ########################
        for label, data in self.nodes(data=True):
            op = data["operator"]
            for in_var in op.inputs:
                if in_var in all_outputs:
                    # link all collected outputs of given variable in inputs field of operator
                    for predecessor in all_outputs[in_var]:
                        # add predecessor output as source; this would also work for non-equal variable names
                        data["inputs"][in_var]["sources"].add(predecessor)
                        # adding into set means there will be no order, but also nothing can be added twice
                        self.add_edge(predecessor, label)
                else:
                    pass  # means, that 'source' will remain an empty list and no incoming edge will be added

        # check for cycle in operator graph
        ###################################
        try:
            find_cycle(self)
        except NetworkXNoCycle:
            pass
        else:
            raise PyRatesException("Found cyclic operator graph. Cycles are not allowed for operators within one node "
                                   "or edge.")

    def __hash__(self):
        return self._h

    def getitem_from_iterator(self, key: str, key_iter: Iterator[str]):
        """
        Helper function for Python magic __getitem__. Accepts an iterator that yields string keys. If `key_iter`
        contains one key, an operator will be (looked for and) returned. If it instead contains two keys, properties of
        a variable that belong to an operator is returned.

        Parameters
        ----------
        key
        key_iter

        Returns
        -------
        item
            operator or variable properties
        """

        try:
            var = next(key_iter)
        except StopIteration:
            # no variable specified, so we return an operator
            item = self.nodes[key]["operator"]
        else:
            # variable specified, so we return variable properties instead
            item = self.nodes[key]["operator"].variables[var]

        return item

    def __iter__(self):
        """Return an iterator containing all operator labels in the operator graph."""
        return iter(self.nodes(data=True))

    def operators(self, get_ops=False, get_vals=False):
        """Alias for self.nodes"""

        if get_ops and get_vals:
            return ((data["label"], op, data["values"]) for op, data in self.nodes(data=True))
        elif get_ops:
            return ((data["label"], op) for op, data in self.nodes(data=True))
        elif get_vals:
            return ((data["label"], data["values"]) for op, data in self.nodes(data=True))
        else:
            return self.nodes


class VectorizedOperatorGraph(DiGraph):
    """Alternate version of `OperatorGraph` that is produced during vectorization. Contents of this version are not
    particularly protected and the instance is not cached."""

    def __init__(self, op_graph: OperatorGraph = None, values: dict = None):

        super().__init__()

        if op_graph is None:
            pass
        else:
            for op_key, data in op_graph:
                try:
                    op = data["operator"]
                except KeyError:
                    # need to add this for the copy capability
                    self.add_operator(op_key, **data)
                else:
                    self.add_operator(op_key,
                                      inputs=data["inputs"],
                                      equations=list(op.equations),
                                      variables=op.variables.to_dict(),
                                      output=op.output)

                # retrieve values from value dict and pass them into variable dictionary of operator
                op_values = values[op_key]  #deepcopy(values[op_key])
                op_vars = self.operators[op_key]["variables"]
                for var_key, value in op_values.items():
                    if op_vars[var_key]["vtype"] == "input_variable":
                        op_vars[var_key]["value"] = value
                        op_vars[var_key]["vtype"] = "input"
                    else:
                        op_vars[var_key]["value"] = value if type(value) is list else [value]

            self.add_edges_from(op_graph.edges)

    def add_operator(self, *args, **kwargs):
        """Alias for `self.add_node`"""
        self.add_node(*args, **kwargs)

    def getitem_from_iterator(self, key: str, key_iter: Iterator[str]):
        """
        Helper function for Python magic __getitem__. Accepts an iterator that yields string keys. If `key_iter`
        contains one key, an operator will be (looked for and) returned. If it instead contains two keys, properties of
        a variable that belong to an operator is returned.

        Parameters
        ----------
        key
        key_iter

        Returns
        -------
        item
            operator or variable properties
        """

        try:
            var = next(key_iter)
        except StopIteration:
            # no variable specified, so we return an operator
            item = self.nodes[key]
        else:
            # variable specified, so we return variable properties instead
            item = self.nodes[key]["variables"][var]

        return item

    @property
    def operators(self):  # , get_ops=False, get_vals=False):
        """Alias for self.nodes"""

        # if get_ops and get_vals:
        #     return ((data["label"], op, data["values"]) for op, data in self.nodes(data=True))
        # elif get_ops:
        #     return ((data["label"], op) for op, data in self.nodes(data=True))
        # elif get_vals:
        #     return ((data["label"], data["values"]) for op, data in self.nodes(data=True))
        # else:
        return self.nodes

    def append_values(self, value_dict: dict) -> dict:
        """Append value along vector dimension of operators.

        Parameters
        ----------
        value_dict

        Returns
        -------
        dict
            Dictionary containing the indices of the appended variable values in the overall vectorized variables.
        """
        var_ranges = {}
        for op_key, variables_updates in value_dict.items():
            original_variables = self.nodes[op_key]["variables"]
            for var_key, value in variables_updates.items():
                var = original_variables[var_key]
                shape = _np.shape(value)
                shape_sum = _np.sum(shape)
                if shape_sum > 1:
                    if _np.all(shape == _np.shape(var["value"])[1:]):
                        raise ValueError(f"Inconsistent dimensions of variable {var}. Dimension of value to add: "
                                         f"{shape} Internal dimension of vectorized value array: "
                                         f"{_np.shape(var['value'])[1:]}")
                    var["value"].append(value)
                elif shape_sum == 0:
                    # case, if shape == ()
                    var["value"].append(value)
                else:
                    var["value"].extend(value)

                # also recompute shape
                old_shape = var["shape"]
                var["shape"] = _np.shape(var["value"])
                var_ranges[(op_key, var_key)] = (old_shape[0], var['shape'][0])

        return var_ranges

    @property
    def var_lengths(self) -> dict:
        """Dictionary containing the length of each variable of each operator.
        """
        var_lengths = {}
        for op in self.operators:
            for var, data in self.nodes[op]['variables'].items():
                var_lengths[(op, var)] = data['shape'][0] if data['shape'] else 1
        return var_lengths
